import json

from . import exceptions
from . import messages


__all__ = (
    'Base',
    'Json'
)


class Base:

    BaseMessage = messages.Base
    RequestMessage = messages.Request
    ResponseMessage = messages.Response
    ErrorMessage = messages.Error

    request_messages = {
        RequestMessage.name: RequestMessage,
    }
    response_messages = {
        ResponseMessage.name: ResponseMessage,
        ErrorMessage.name: ErrorMessage,
    }

    def decode(self, raw_data):
        raise NotImplementedError

    def encode(self, raw_data):
        raise NotImplementedError

    def decode_request(self, raw_request):
        request = self.decode(raw_request)
        base_message = self.BaseMessage(**request)
        message_cls = self.request_messages.get(base_message['name'])
        if message_cls is None:
            raise exceptions.MessageError(
                {'name': 'Name {} not allowed'.format(base_message['name'])},
            )
        else:
            return message_cls(**request)

    def encode_response(self, response_message):
        assert response_message['name'] in self.response_messages
        return  self.encode(dict(response_message))


class Json(Base):

    def decode(self, raw_data):
        try:
            python_data = json.loads(raw_data)
        except json.decoder.JSONDecodeError as exc:
            raise exceptions.JSONDecodeError(exc.lineno, exc.colno, exc.pos)
        if not isinstance(python_data, dict):
            raise exceptions.RequestTypeError('dict', type(python_data).__name__)
        return python_data

    def encode(self, python_data):
        assert isinstance(python_data, dict), 'Response must be the dict'
        return json.dumps(python_data)




import os
import asyncio
import logging
import mimetypes
import collections
import email.message
import websockets
import websockets.http
import websockets.handshake
import websockets.protocol


logger = logging.getLogger(__name__)


class HTTPResponse:
    default_content_type = 'text/html'
    default_charset = 'UTF-8'
    default_status = '200 OK'

    def __init__(self, body, status=None, content_type=None, charset=None):
        self.body = body
        self.status = status or self.default_status
        self.content_length = str(len(body))
        self.content_type = content_type or self.default_content_type
        if content_type.startswith('text/'):
            charset = charset or self.default_charset
            self.content_type += ';charset={}'.format(charset)

    def __str__(self):
        headers = [
            ('Content-Type', self.content_type),
            ('Content-Length', self.content_length),
        ]
        response = ['HTTP/1.1 ', self.status, '\r\n']
        for key, value in headers:
            response.append('{}: {}\r\n'.format(key, value))
        response.append('\r\n')
        response.append(self.body)
        return ''.join(response)


def try_files(root, path):
    def translate(root, path):
        parts = filter(None, path.split('/'))
        for part in parts:
            drive, part = os.path.splitdrive(part)
            head, part = os.path.split(part)
            if drive or head or part in [os.curdir, os.pardir]:
                return None
        return os.path.join(root, path)
    filename = translate(root, path)
    if filename is None or not os.path.exists(filename):
        body = 'File not found'
        return HTTPResponse(body, status='404 Not Found')
    with open(filename) as fp:
        content_type, charset = mimetypes.guess_type(filename)
        body = fp.read()
        return HTTPResponse(body, content_type=content_type, charset=charset)


class TryFilesError(Exception):
    def __init__(self, path):
        self.path = path


class WebSocketServerProtocol(websockets.WebSocketServerProtocol):

    @asyncio.coroutine
    def handshake(self, origins=None, subprotocols=None, extra_headers=None):
        # Read handshake request.
        try:
            path, headers = yield from websockets.http.read_request(self.reader)
        except Exception as exc:
            raise websockets.InvalidHandshake("Malformed HTTP message") from exc

        self.request_headers = headers
        self.raw_request_headers = list(headers.raw_items())

        get_header = lambda k: headers.get(k, '')

        try:
            key = websockets.handshake.check_request(get_header)
        except websockets.InvalidHandshake as exc:
            raise TryFilesError(path)

        if origins is not None:
            origin = get_header('Origin')
            if not set(origin.split() or ['']) <= set(origins):
                raise websockets.InvalidOrigin("Origin not allowed: {}".format(origin))

        if subprotocols is not None:
            protocol = get_header('Sec-WebSocket-Protocol')
            if protocol:
                client_subprotocols = [p.strip() for p in protocol.split(',')]
                self.subprotocol = self.select_subprotocol(
                    client_subprotocols, subprotocols)

        headers = []
        set_header = lambda k, v: headers.append((k, v))
        set_header('Server', websockets.http.USER_AGENT)
        if self.subprotocol:
            set_header('Sec-WebSocket-Protocol', self.subprotocol)
        if extra_headers is not None:
            if callable(extra_headers):
                extra_headers = extra_headers(path, self.raw_request_headers)
            if isinstance(extra_headers, collections.abc.Mapping):
                extra_headers = extra_headers.items()
            for name, value in extra_headers:
                set_header(name, value)
        websockets.handshake.build_response(set_header, key)

        self.response_headers = email.message.Message()
        for name, value in headers:
            self.response_headers[name] = value
        self.raw_response_headers = headers

        # Send handshake response. Since the status line and headers only
        # contain ASCII characters, we can keep this simple.
        response = ['HTTP/1.1 101 Switching Protocols']
        response.extend('{}: {}'.format(k, v) for k, v in headers)
        response.append('\r\n')
        response = '\r\n'.join(response).encode()
        self.writer.write(response)

        assert self.state == websockets.protocol.CONNECTING
        self.state = websockets.protocol.OPEN
        self.opening_handshake.set_result(True)

        return path

    @asyncio.coroutine
    def handler(self):
        # Since this method doesn't have a caller able to handle exceptions,
        # it attemps to log relevant ones and close the connection properly.
        try:

            try:
                path = yield from self.handshake(
                    origins=self.origins, subprotocols=self.subprotocols,
                    extra_headers=self.extra_headers)
            except Exception as exc:
                logger.info("Exception in opening handshake: {}".format(exc))
                if isinstance(exc, TryFilesError):
                    response = try_files(exc.path)
                elif isinstance(exc, websockets.InvalidOrigin):
                    response = 'HTTP/1.1 403 Forbidden\r\n\r\n' + str(exc)
                elif isinstance(exc, websockets.InvalidHandshake):
                    response = 'HTTP/1.1 400 Bad Request\r\n\r\n' + str(exc)
                    import pdb;pdb.set_trace()
                else:
                    response = ('HTTP/1.1 500 Internal Server Error\r\n\r\n'
                                'See server log for more information.')
                self.writer.write(response.encode())
                raise

            try:
                yield from self.ws_handler(self, path)
            except Exception:
                logger.error("Exception in connection handler", exc_info=True)
                yield from self.fail_connection(1011)
                raise

            try:
                yield from self.close()
            except Exception as exc:
                logger.info("Exception in closing handshake: {}".format(exc))
                raise

        except Exception:
            # Last-ditch attempt to avoid leaking connections on errors.
            try:
                self.writer.close()
            except Exception:                               # pragma: no cover
                pass

        finally:
            # Unregister the connection with the server when the handler task
            # terminates. Registration is tied to the lifecycle of the handler
            # task because the server waits for tasks attached to registered
            # connections before terminating.
            self.ws_server.unregister(self)



import os

import jinja2

from ikcms.utils import cached_property
from ..base import RenderComponent
from . import extensions


class Jinja2Component(RenderComponent):

    filters = {}
    tests = {}
    globals = {}
    extensions = [
        extensions.CacheTag,
        extensions.I18n,
        extensions.ShowTag,
        extensions.Macros,
    ]
    autoescape = True
    default_file_ext = 'html'

    def __init__(self, app):
        super(Jinja2Component, self).__init__(app)
        self.paths = [self.app.cfg.dirpath(path) for path in self.paths]
        self._env = self._make_env()

    @cached_property
    def paths(self):
        return ['{}/templates'.format(self.app.cfg.SITE_DIR)]

    def resolve(self, name):
        base, ext = os.path.splitext(name)
        if not ext:
            return '.'.join((base, self.default_file_ext))
        return name

    def _make_env(self):
        loaders = []
        for path in self.paths:
            if path.scheme == '':
                loaders.append(jinja2.FileSystemLoader(path.url.path))
            elif path.scheme == 'pkg':
                loaders.append(
                    jinja2.PackageLoader(path.url.package, path.url.path),
                )
        env = jinja2.Environment(
            loader=jinja2.ChoiceLoader(loaders),
            autoescape=self.autoescape,
            extensions=self.extensions,
        )
        env.filters.update(self.filters)
        env.tests.update(self.tests)
        env.globals.update(self.globals)
        return env

    def render(self, template_name, **context):
        context = context or {}
        template_name = self.resolve(template_name)
        return self._env.get_template(template_name).render(**context)
    __call__ = render




component = Jinja2Component.create_cls



import os
from collections import OrderedDict

from ikcms.utils import cached_property
import ikcms.apps.composite


class Cfg(ikcms.apps.composite.Cfg):
    SITE_ID = 'admin'

    DATABASES = {
    }

    STATIC_ENABLED = False

    MODEL_LOCK_TIMEOUT = 5 * 60
    MODEL_LOCK_RENEW = 60

    DATABASE_PARAMS = {
        'pool_size': 10,
        'max_overflow': 50,
        'pool_recycle': 3600,
    }

    @cached_property
    def CMS_STATIC_DIR(self):
        return os.path.join(self.STATIC_DIR, 'cms')

    CMS_STATIC_URL = '/cms-static/'

    TEMPLATE_IMPORT_SETTINGS = ['STATIC_URL', 'CMS_STATIC_URL']

    @cached_property
    def MANIFESTS(self):
        return OrderedDict([
            ("cms", {
                "path": self.CMS_STATIC_DIR,
                "url": self.CMS_STATIC_URL,
                "css": 'css/Manifest',
                "js": 'js/Manifest'
            }),
            ("", {
                "path": self.STATIC_DIR,
                "url": self.STATIC_URL,
                "css": 'css/Manifest',
                "js": 'js/Manifest'
            }),
        ])

    @cached_property
    def FORM_TEMP(self):
        return os.path.join(self.ROOT_DIR, 'form-temp')

    FORM_TEMP_URL = '/form-temp/'

    @cached_property
    def MEDIA_ROOT_ADMIN(self):
        return os.path.join(self.ROOT_DIR, 'media', 'admin')

    @cached_property
    def MEDIA_ROOT_FRONT(self):
        return os.path.join(self.ROOT_DIR, 'media', 'front')

    MEDIA_URL = '/media/'
    
    PREVIEW_STATIC_URL = '/preview/static/'

    REDIS_DB = 1



import os
import sys


def manage(apps, paths=[]):
    for path in paths:
        sys.path.insert(0, os.path.abspath(path))

    import iktomi.cli
    import iktomi.cli.lazy

    commands = {}

    for name in apps:
        module = __import__(name)
        App = module.App
        Cfg = module.Cfg
        for prefix, cli in getattr(App, 'commands', {}).items():
            commands[prefix] = iktomi.cli.lazy.LazyCli(
                lambda cli=cli, App=App, Cfg=Cfg: cli(App, Cfg),
            )

    iktomi.cli.manage(commands)



from unittest import TestCase, skipIf
from unittest.mock import MagicMock

from sqlalchemy import sql
from ikcms.components.db.sqla import component

from tests.cfg import cfg
from tests.models import create_models1, create_models2


try:
    import MySQLdb
    mysql_skip = False
except ImportError:
    mysql_skip = True

try:
    raise ImportError
    import psycopg2
    pg_skip = False
except ImportError:
    pg_skip = True


class SQLAComponentTestCase(TestCase):

    test_models = {
        'db1': create_models1(),
        'db2': create_models2(),
    }

    @skipIf(mysql_skip, 'MySQLdb not installed')
    def test_mysql(self):
        self._db_test(cfg.MYSQL_URL)

    @skipIf(pg_skip, 'Psycopg2 not instaled')
    def test_postgress(self):
        self._db_test(cfg.POSTGRESS_URL)

    def _db_test(self, db_url):
        app = MagicMock()
        del app.db
        app.cfg.DATABASES = {
            'db1': db_url,
            'db2': db_url,
        }

        def get_models(db_id):
            self.assertIn(db_id, app.cfg.DATABASES)
            return self.test_models[db_id]

        Component = component(get_models=get_models)

        db = Component.create(app)
        self.assertEqual(db.models, self.test_models)
        engine1 = db.engines['db1']
        engine2 = db.engines['db2']
        self.assertEqual(db.binds[self.test_models['db1'].test_table1], engine1)
        self.assertEqual(db.binds[self.test_models['db2'].test_table2], engine2)
        test_table1 = self.test_models['db1'].test_table1

        session = db()
        db.reset_all()
        Test = db.models['db1'].Test
        test = Test(id=5, title='test_title')
        session.add(test)
        session.commit()
        test = session.query(Test).filter_by(id=5).one()
        self.assertEqual(test.id, 5)
        self.assertEqual(test.title, 'test_title')

        test.title = 'test_title2'
        session.commit()
        cnt = session.query(Test).filter_by(id=5, title='test_title2').count()
        self.assertEqual(cnt, 1)

        session.delete(test)
        session.commit()
        cnt = session.query(Test).filter_by(id=5, title='test_title2').count()
        self.assertEqual(cnt, 0)
        session.close()




# -*- coding: utf-8 -*-
from datetime import datetime

from sqlalchemy import Column, Integer, String, DateTime, Boolean, ForeignKey
from sqlalchemy.orm import relationship, object_session

from iktomi.db.sqla.types import StringList
from iktomi.auth import encrypt_password, check_password as check_password_


__all__ = ['AdminUser', 'AdminGroup', 'AdminUser_AdminGroup']


def AdminUser(models):

    id = Column(Integer, primary_key=True)
    name = Column(String(250), nullable=False, default='')
    login = Column(String(32), nullable=False, unique=True)
    email = Column(String(200), index=True)
    password = Column(String(250), nullable=False)
    creation_time = Column(DateTime, default=datetime.now, nullable=False)
    # roles = Column(StringList(250), nullable=False, default=[])
    active = Column(Boolean, default=True)

    groups = relationship(models.AdminGroup,
                          secondary=lambda: models.AdminUser_AdminGroup.__table__)

    def __unicode__(self):
        if self.id is None:
            return u'Новый редактор'
        return u'Редактор: %s (%s)' % (self.name, self.login)

    def set_password(self, password):
        self.password = encrypt_password(password)

    def check_password(self, password):
        return check_password_(password, self.password)

    def add_to_group(self, group_title):
        db = object_session(self)
        group_object = db.query(models.AdminGroup) \
            .filter_by(title=group_title).first()
        self.groups.append(group_object)
        db.commit()

    @property
    def roles(self):
        # This is needed for backward compartibility
        return [group.title for group in self.groups]

    @roles.setter
    def roles(self, value):
        for group_title in value:
            if group_title not in self.roles:
                self.add_to_group(group_title)

    @property
    def can_publish(self):
        return 'publisher' in self.roles


def AdminGroup(models):

    id = Column(Integer, primary_key=True)
    title = Column(String(250), nullable=False, unique=True)
    order = Column(Integer, nullable=False, default=0)

    __mapper_args__ = {
        'order_by': order.asc()
    }

    @property
    def ru_title(self):
        from admin.streams.admins import ROLES
        return dict(ROLES).get(self.title)


def AdminUser_AdminGroup(models):

    user_id = Column(Integer, ForeignKey('AdminUser.id'),
                     nullable=False, primary_key=True)
    user = relationship('AdminUser')
    group_id = Column(Integer, ForeignKey('AdminGroup.id'),
                      nullable=False, primary_key=True)
    group = relationship('AdminGroup')




import logging
from importlib import import_module
from ikcms.cli.base import Cli


class SphinxCli(Cli):
    name = 'sphinx'

    def command_xmlpipe(self, name):
        app = self.create_app()
        xmlpipes = app.sphinx.xmlpipes
        import_string = xmlpipes.get(name)
        if not import_string:
            logging.info('Unknown xmlpipe "%s"', name)
            raise SystemExit
        module_name, xmlpipe_name = import_string.rsplit('.', 1)

        try:
            module = import_module(module_name)
        except ImportError as exc:
            logging.info('Could not load module "%s"', module_name)
            raise SystemExit

        try:
            xmlpipe = getattr(module, xmlpipe_name)
        except AttributeError as exc:
            logging.info('Could not find function "%s" in "%s"',
                         xmlpipe_name,
                         module_name)
            raise SystemExit

        xmlpipe(app)



import os

from iktomi.utils import cached_property

import ikcms.apps.composite


class Cfg(ikcms.apps.composite.Cfg):

    STATIC_ENABLED = True

    @cached_property
    def STATIC_DIR(self):
        return os.path.join(self.SITE_DIR, 'static')

    STATIC_URL = '/static/'

    @cached_property
    def MEDIA_DIR(self):
        return os.path.join(self.ROOT_DIR, 'media')

    MEDIA_URL = '/media/'

    DATABASES = {'': 'mysql://root@localhost/web?charset=utf8'}
    DATABASE_PARAMS = {}

    @cached_property
    def MEDIA_DIR(self):
        return os.path.join(self.ROOT_DIR, 'media')



import sqlalchemy as sa
from iktomi.utils import cached_property

from .query import Query, PubQuery
from . import exc
from . import relations


__all__ = [
    'Registry',
    'Core',
    'I18nMixin',
    'PublicationMixin',
    'Base',
    'I18n',
    'Pub',
    'I18nPub',
]


class Registry(dict):

    def __init__(self, metadata):
        super().__init__()
        self.metadata = metadata
        for key in metadata:
            self[key] = {}
        self.create_schema_mappers = []

    def create_schema(self):
        mappers = []
        for mapper in self.create_schema_mappers:
            mapper.create_table()

        for mapper in self.create_schema_mappers:
            for relation in mapper.relations.values():
                relation.create_tables()

    def set_mapper(self, path, mapper):
        if isinstance(path, str):
            parts = path.split('.')
        elif isinstance(path, [list, tuple]):
            parts = path
        #XXX
        parts, name = parts[:-1], parts[-1]
        d = self
        for part in parts:
            d = d.setdefault(part, {})
        d[name] = mapper

    def get_mapper(self, path):
        if isinstance(path, str):
            parts = path.split('.')
        elif isinstance(path, [list, tuple]):
            parts = path
        #XXX
        parts, name = parts[:-1], parts[-1]
        d = self
        for part in parts:
            d = d[part]
        return d[name]


    @classmethod
    def from_db_ids(cls, db_ids):
        return cls({db_id: sa.MetaData() for db_id in db_ids})


class Core:

    name = None
    query_class = Query

    def __init__(self, registry, db_id):
        assert self.name
        self.registry = registry
        self.db_id = db_id

    @cached_property
    def id(self):
        return '.'.join([self.db_id, self.name])

    @property
    def tablename(self):
        return self.name

    @cached_property
    def table(self):
        return self.registry.metadata[self.db_id].tables[self.tablename]

    @cached_property
    def relations(self):
        return self.create_relations()

    def get_mapper(self, **kwargs):
        db_id = kwargs.get('db_id', self.db_id)
        name = kwargs.get('name', self.name)
        return self.registry[db_id][name]

    def create_table(self):
        return sa.Table(
            self.tablename,
            self.registry.metadata[self.db_id],
            *(self.create_system_columns() + self.create_columns())
        )

    def create_id_column(self):
        return sa.Column('id', sa.Integer, primary_key=True, autoincrement=True)

    def create_system_columns(self):
        return [self.create_id_column()]

    def create_columns(self):
        return []

    def create_relations(self):
        return {}

    async def schema_initialize(self, session):
        pass

    @property
    def c(self):
        return dict(self.table.c, **self.relations)

    @property
    def table_keys(self):
        return set(self.table.c.keys())

    @property
    def relation_keys(self):
        return set(self.relations.keys())

    @property
    def allowed_keys(self):
        return self.table_keys.union(self.relation_keys)

    def query(self):
        return self.query_class(self)

    def div_keys(self, keys=None):
        if keys is None:
            keys = self.allowed_keys
        else:
            keys = set(keys)
            error_keys = keys - self.allowed_keys
            assert not error_keys, 'Unknown fields {}'.format(error_keys)
        table_keys = keys.intersection(self.table_keys)
        relation_keys = keys.intersection(self.relation_keys)
        return table_keys, relation_keys

    async def select_items(self, session, query, keys=None):
        ids = await self._select_ids(session, query)
        if ids:
            return await self._load_items(session, ids, keys)
        else:
            return []

    async def select_first_item(self, session, query, keys=None):
        items = await self.select_items(session, query.limit(1), keys)
        return items and items[0] or None

    async def insert_item(self, session, values, keys=None):
        keys = keys or list(values.keys())
        assert set(keys).issubset(self.allowed_keys), \
            'Keys {} not allowed. Allowed keys={}'.format(
                set(keys).difference(self.allowed_keys), self.allowed_keys)
        table_keys, relation_keys = self.div_keys(keys)
        table_values = {key: values[key] for key in table_keys}
        relation_values = {key: values[key] for key in relation_keys}

        # set defaults, aio engine don't do it
        for key, column in self.table.c.items():
            if key not in table_values and column.default is not None:
                if column.default.is_callable:
                    table_values[key] = column.default.arg(self)
                else:
                    table_values[key] = column.default.arg
        query = sa.sql.insert(self.table).values(table_values)
        result = await session.execute(query)
        item = dict(values)
        if values.get('id') is None:
            item['id'] = result.lastrowid
        # store relations
        for key in relation_keys:
            await self.relations[key].store(session, item['id'], values[key])
        return item

    async def update_item(self, session, query, item_id, values, keys=None):
        await self._exists_check(session, query, item_id)
        return await self.update_item_by_id(session, item_id, values, keys)

    async def update_item_by_id(self, session, item_id, values, keys=None):
        keys = keys or list(values.keys())
        table_keys, relation_keys = self.div_keys(keys)
        table_values = {key: values[key] for key in table_keys}
        relation_values = {key: values[key] for key in relation_keys}
        query = sa.sql.update(self.table).values(**table_values)
        query = query.where(self.c['id'] == item_id)
        result = await session.execute(query)
        item_id = values.get('id', item_id)
        for key, value in relation_values.items():
            await self.relations[key].store(session, item_id, value)
        return dict(values, id=item_id)

    async def delete_item(self, session, query, item_id):
        await self._exists_check(session, query, item_id)
        await self.delete_item_by_id(session, item_id)

    async def delete_item_by_id(self, session, item_id):
        query = sa.sql.delete(self.table).where(self.c['id'] == item_id)
        for key in self.relation_keys:
            await self.relations[key].delete(session, item_id)
        await session.execute(query)

    async def count_items(self, session, query):
        query = query.with_only_columns([sa.func.count(self.c['id'])])
        result = await session.execute(query)
        row = await result.fetchone()
        return row[0]

    async def _select_ids(self, session, query):
        rows = list(await session.execute(query))
        return [row[self.c['id']] for row in rows]

    async def _load_items(self, session, ids, keys=None):
        table_keys, relation_keys = self.div_keys(keys)
        table_keys.add('id')
        query = sa.sql.select([self.c[key] for key in table_keys])
        if len(ids) == 1:
            query = query.where(self.c['id'] == ids[0])
        else:
            query = query.where(self.c['id'].in_(ids))
        rows = list(await session.execute(query))
        items = [dict(row) for row in rows]
        item_by_id = {item['id']: item for item in items}
        sorted_items = [item_by_id[id] for id in ids]\
        # load relations
        for key in relation_keys:
            result = await self.relations[key].load(session, item_by_id.keys())
            for id, value in result.items():
                item_by_id[id][key] = value
        return sorted_items

    async def _exists_check(self, session, query, item_id):
        ids = await self._select_ids(session, query.filter_by(id=item_id))
        if len(ids) == 0:
            raise exc.ItemNotFoundError(item_id)
        if len(ids) > 1:
            raise exc.OrmError('There are many items with id={}'.format(item_id))

    #factory
    @classmethod
    def create_mappers(cls, registry, **kwargs):
        return [cls(registry, **kwargs)]

    @classmethod
    def register_mappers(cls, registry, mappers, create_schema=True):
        for mapper in mappers:
            registry.set_mapper(mapper.id, mapper)
            if create_schema:
                registry.create_schema_mappers.append(mapper)

    @classmethod
    def create(cls, registry, create_schema=True, **kwargs):
        mappers = cls.create_mappers(registry, **kwargs)
        cls.register_mappers(registry, mappers, create_schema)
        return mappers

    @classmethod
    def from_model(cls, registry, models, name=None):
        for model in models:
            mapper_cls = type('ModelMapper', (cls,), {
                'name': name or model.__name__,
                'model': model,
                'create_relations': lambda self: get_model_relations(self),
            })

            mapper_cls.create(
                registry,
                db_id=get_model_db_id(registry, model),
                create_schema=False,
            )

    async def fill(self, session, query, data, str_path):
        paths = str_path.split('.')
        if isinstance(data, dict):
            data = [data]
        while paths:
            tp = self._get_list_type(data)
            if tp == list:
                tmp_data = []
                for item in data:
                    tmp_data += item
                data = tmp_data
            elif tp == dict:
                path = paths.pop(0)
                prev_data = data
                data = [x[path] for x in data]
            else:
                raise TypeError(tp)

        tp = self._get_list_type(data)
        if tp == list:
            await self._fill_lists(session, query, data)
        else:
            await self._fill_dicts(session, query, prev_data, path)

    def _get_list_type(self, items_list):
        data = [item for item in items_list if item]
        tps = set([type(item) for item in items_list])
        if len(tps) > 1:
            raise TypeError(tps)
        return tps.pop()

    async def _fill_lists(self, session, query, data):
        ids = set()
        for items_list in data:
            ids.update(items_list)
        items = await query.id(ids).select_items(session)
        items_by_id = {item['id']: item for item in items}
        for items_list in data:
            items = [items_by_id[x] for x in items_list]
            items_list.clear()
            items_list.extend(items)

    async def _fill_dicts(self, session, query, data, key):
        ids = set()
        for items_dict in data:
            ids.add(items_dict[key])
        items = await query.id(ids).select_items(session)
        items_by_id = {id: item for item in items}
        for items_dict in data:
            items_dict[key] = items_by_id[items_dict[key]]


class I18nMixin:

    STATE_ABSENT = 'absent'
    STATE_NORMAL = 'normal'

    langs = ['ru', 'en']
    common_keys = []

    def __init__(self, registry, lang, **kwargs):
        assert lang in self.langs
        super().__init__(registry, **kwargs)
        self.lang = lang

    @cached_property
    def id(self):
        return '.'.join([self.db_id, self.lang, self.name])

    # schema
    def create_id_column(self):
        if self.lang == self.langs[0]:
            return sa.Column(
                'id', sa.Integer, primary_key=True, autoincrement=True)
        else:
            return sa.Column(
                'id',
                sa.Integer,
                sa.ForeignKey(
                    self.get_mapper(lang=self.langs[0]).c['id'],
                    ondelete='cascade',
                ),
                primary_key=True,
                autoincrement=False,
            )

    def create_state_column(self):
        return sa.Column(
            'state',
            sa.Enum(*self.get_states()),
            nullable=False,
        )

    def create_system_columns(self):
        return [
            self.create_id_column(),
            self.create_state_column(),
        ]

    @property
    def tablename(self):
        return '{}{}'.format(self.name, self.lang.capitalize())

    #i18n methods
    def get_mapper(self, **kwargs):
        db_id = kwargs.get('db_id', self.db_id)
        lang = kwargs.get('lang', self.lang)
        name = kwargs.get('name', self.name)
        return self.registry[db_id][lang][name]

    def get_states(self):
        states = getattr(super(), 'get_states', lambda: set())()
        states.add(self.STATE_NORMAL)
        states.add(self.STATE_ABSENT)
        return states

    def query(self):
        return self.i18n_base_query().where(self.c['state'] != self.STATE_ABSENT)

    def absent_query(self):
        return self.i18n_base_query().filter_by(state=self.STATE_ABSENT)

    async def insert_item(self, session, values, keys=None):
        results = {}
        values = dict(values)
        for lang in self.langs:
            mapper = self.i18n_mappers[lang]
            if lang == self.lang:
                lang_values = dict(values)
                self.set_normal_state(lang_values)
                lang_keys = keys
            else:
                lang_values = dict(values)
                self.set_absent_state(lang_values)
                if keys is None:
                    common_keys = set(self.common_keys)
                else:
                    common_keys = set(self.common_keys).intersection(keys)
                lang_keys = common_keys.union({'id', 'state'})
            insert_item = mapper.i18n_base_insert_item
            results[lang] = await insert_item(session, lang_values, lang_keys)
            values['id'] = results[lang]['id']
        return results[self.lang]

    async def update_item_by_id(self, session, item_id, values, keys=None):
        assert 'id' not in values or values['id'] == item_id,\
            'Changing item_id not permitted'
        if keys is None:
            common_keys = set(self.common_keys)
        else:
            common_keys = set(self.common_keys).intersection(keys)
        for lang in self.langs:
            mapper = self.i18n_mappers[lang]
            update_item = mapper.i18n_base_update_item_by_id
            if lang == self.lang:
                result = await update_item(session, item_id, values, keys)
            elif common_keys:
                await update_item(
                    session,
                    item_id,
                    values,
                    common_keys,
                )
        return result

    async def delete_item_by_id(self, session, item_id):
        for lang in self.langs[::-1]:
            mapper = self.i18n_mappers[lang]
            await mapper.i18n_base_delete_item_by_id(session, item_id)

    def set_normal_state(self, values):
        values['state'] = self.STATE_NORMAL

    def set_absent_state(self, values):
        values['state'] = self.STATE_ABSENT

    def i18n_base_query(self):
        return super().query()

    @cached_property
    def i18n_mappers(self):
        return {lang: self.get_mapper(lang=lang) for lang in self.langs}

    async def i18n_base_insert_item(self, session, values, keys=None):
        return await super().insert_item(session, values, keys)

    async def i18n_base_update_item_by_id(self, session, item_id, values,
                                          keys=None):
        return await super().update_item_by_id(session, item_id, values, keys)

    async def i18n_base_delete_item_by_id(self, session, item_id):
        return await super().delete_item_by_id(session, item_id)

    async def i18n_get_version(self, session, item_id, keys=None):
        query = self.i18n_mappers[self.lang].i18n_base_query().id(item_id)
        await self._exists_check(session, query, item_id)
        items = await query.select_items(session, keys)
        return items[0]

    async def i18n_create_version(self, session, item_id):
        query = self.absent_query()
        values = {}
        self.set_normal_state(values)
        await self.update_item(session, query, item_id, values, keys=['state'])

    #factory
    @classmethod
    def create_mappers(cls, registry, **kwargs):
        return [cls(registry, lang=lang, **kwargs) for lang in cls.langs]


class PublicationMixin:

    query_class = PubQuery

    STATE_PRIVATE = 'private'
    STATE_PUBLIC = 'public'

    STATE_NORMAL = STATE_PRIVATE

    db_ids = ['admin', 'front']

    def get_states(self):
        states = getattr(super(), 'get_states', lambda: set())()
        states.add(self.STATE_PRIVATE)
        states.add(self.STATE_PUBLIC)
        return states

    # schema
    def create_state_column(self):
        return sa.Column(
            'state',
            sa.Enum(*self.get_states()),
            nullable=False,
        )

    def create_system_columns(self):
        return [
            self.create_id_column(),
            self.create_state_column(),
        ]

    # pub methods
    def pub_base_query(self):
        return super().query()

    def query(self):
        return self.pub_base_query()

    def private_query(self):
        return self.pub_base_query().filter_by(state=self.STATE_PRIVATE)

    def public_query(self):
        return self.pub_base_query().filter_by(state=self.STATE_PUBLIC)

    @cached_property
    def pub_mappers(self):
        return [self.get_mapper(db_id=db_id) for db_id in self.db_ids]

    async def insert_item(self, session, values, keys=None):
        assert self.db_id == self.db_ids[0], \
            'Insert denied for "{}" mapper'.format(self.db_id)
        values = dict(values)
        insert_item = self.pub_base_insert_item
        self.set_private_state(values)
        item = await insert_item(session, values, keys=keys)
        values = {'id': item['id']}
        self.set_private_state(values)
        insert_item = self.pub_mappers[1].pub_base_insert_item
        await insert_item(session, values, keys={'id', 'state'})
        return item

    async def delete_item(self, session, query, item_id):
        assert self.db_id == self.db_ids[0], \
            'Delete denied for "{}" mapper'.format(self.db_id)
        await super().delete_item(session, query, item_id)

    async def delete_item_by_id(self, session, item_id):
        for mapper in self.pub_mappers:
            await mapper.pub_base_delete_item_by_id(session, item_id)

    async def pub_base_delete_item_by_id(self, session, item_id):
        await super().delete_item_by_id(session, item_id)

    async def publish(self, session, query, item_id):
        assert self.db_id == self.db_ids[0], \
            'Publish denied for "{}" mapper'.format(self.db_id)
        assert self.db_id == self.db_ids[0]
        await self._exists_check(session, query, item_id)
        admin_mapper = self.pub_mappers[0]
        items = await admin_mapper.select_items(session, query.id(item_id))
        admin_item = items[0]
        assert admin_item['state'] == 'private', \
            'Item id="{}" is not private'.format(admin_item['id'])
        values = admin_item
        self.set_public_state(admin_item)
        await admin_mapper.update_item_by_id(session, item_id, values, ['state'])
        front_mapper = self.pub_mappers[1]
        await front_mapper.update_item_by_id(
            session,
            item_id,
            values,
        )

    def set_private_state(self, values):
        values['state'] = self.STATE_PRIVATE

    def set_public_state(self, values):
        values['state'] = self.STATE_PUBLIC

    async def pub_base_insert_item(self, session, values, keys=None):
        return await super().insert_item(session, values, keys)

    async def pub_base_delete_item(self, session, query, item_id):
        return await super().delete_item(session, query, item_id)

    #factory
    @classmethod
    def create_mappers(cls, registry, **kwargs):
        result = []
        for db_id in cls.db_ids:
            mappers = super().create_mappers(registry, db_id=db_id, **kwargs)
            result.extend(mappers)
        return result


class MarkDeletedMixin:

    STATE_NORMAL = 'normal'
    STATE_DELETED = 'deleted'

    async def delete_item_by_id(self, session, item_id):
        for mapper in self.internal_core.internal_mappers:
            mapper.update_item_by_id(
                session,
                item_id,
                {'state': self.DELETED_STATE},
                ['state'],
            )

    @classmethod
    def get_states(cls):
        states = getattr(super(), 'get_states', lambda: {})()
        states.add(cls.STATE_NORMAL)
        states.add(cls.STATE_DELETED)
        return states


class Base(Core):
    pass

class I18n(I18nMixin, Base):
    pass

class Pub(PublicationMixin, Base):
    pass

class I18nPub(PublicationMixin, I18nMixin, Base):
    pass


class I18nPubMD(
        PublicationMixin,
        I18nMixin,
        MarkDeletedMixin,
        Core,
):
    pass


def get_model_db_id(registry, model):
    model_meta = model.__table__.metadata
    for db_id, meta in registry.metadata.items():
        if meta == model_meta:
            return db_id


def get_model_relations(mapper):
    relations_dict = {}
    for name in dir(mapper.model):
        attr = getattr(mapper.model, name)
        if isinstance(attr, sa.orm.attributes.InstrumentedAttribute):
            prop = attr.property
            if isinstance(prop, sa.orm.ColumnProperty):
                pass
            elif isinstance(prop, sa.orm.RelationshipProperty):
                if prop.secondaryjoin is None:
                    pass
                else:
                    local1, remote1 = prop.local_remote_pairs[0]
                    local2, remote2 = prop.local_remote_pairs[1]
                    order_field = remote1.table.c.get('order')
                    # argument is class
                    remote_name = getattr(prop.argument, '__name__', None)
                    # argument is classname
                    if remote_name is None:
                        remote_name = getattr(prop.argument, 'arg', None)
                    relations_dict[prop.key] = relations.M2M(
                        mapper,
                        remote_name,
                        ordered='order' in remote1.table.c,
                    )
                return {}
    return relations_dict




from iktomi.utils import cached_property
from .i18n import N_
from .i18n import M_
from .paths import FilePath
from .paths import DirPath
from .paginator import Paginator
from .paginator import PageNotFound



# -*- coding: utf-8 -*-
import time
from datetime import datetime

from sqlalchemy import types
from sqlalchemy.sql import expression

__all__ = ['Float', 'String', 'Integer', 'Timestamp', 'Multi', 'Boolean']


class _TypeDecorator(types.TypeDecorator):
    default = None

    def process_bind_param(self, value, dialect):
        return self.default if value is None else value


class Float(_TypeDecorator):
    impl = types.Float
    default = 0.0


class String(_TypeDecorator):
    impl = types.String
    default = u''


class Integer(_TypeDecorator):
    impl = types.Integer
    default = 0


class Timestamp(types.TypeDecorator):
    impl = types.Integer

    def process_bind_param(self, value, dialect):
        if value is not None:
            return int(time.mktime(value.timetuple()))
        return 0

    def process_result_value(self, value, dialect):
        if value:
            return datetime.fromtimestamp(value) if value else None
        return None


class Boolean(types.TypeDecorator):
    impl = types.Boolean

    def process_bind_param(self, value, dialect):
        return int(value)

    def process_result_value(self, value, dialect):
        return bool(value)


class Multi(types.UserDefinedType):
    type = types.Integer


class MultiExpression(expression.Tuple):
    __visit_name__ = 'multi'
    type = Multi

    def __init__(self, clauses, **kw):
        super(MultiExpression, self).__init__(*clauses, **kw)

    def self_group(self, against=None):
        return self



from ikcms.forms import fields


__all__ = (
    'item_id',
    'filters',
    'page',
    'page_size',
    'order',
    'kwargs',
    'values',
)

class item_id(fields.Int):
    name = 'item_id'
    label = 'Идентификатор документа'
    required = True


class filters(fields.RawDict):
    name = 'filters'
    label = 'Словарь фильтров'
    to_python_default = {}
    raw_required = False


class page(fields.Int):
    name = 'page'
    label = 'Номер страницы'
    to_python_default = 1
    min_value = 1
    raw_required = False


class page_size(fields.Int):
    name = 'page_size'
    label = 'Размер страницы'
    to_python_default = 1
    min_value = 1
    raw_required = False


class order(fields.String):
    name = 'order'
    label = 'Сортировка'
    to_python_default = '+id'
    required = True
    regex = r'[+\-]{1}'
    regex_error = 'Order value must startswith "+" or "-"',
    raw_required = False


class kwargs(fields.RawDict):
    name = 'kwargs'
    label = 'Ключевые аргументы'
    to_python_default = {}
    raw_required = False


class values(fields.RawDict):
    name = 'values'
    label = 'Значения'
    raw_required = True



from sqlalchemy import func
from sqlalchemy import Column
from sqlalchemy.orm import column_property
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import DeclarativeMeta
from . import types


class ColumnDecorator(Column):
    sphinx_attribute = True
    column_type = None
    def __init__(self, *args, **kwargs):
        super(ColumnDecorator, self).__init__(self.column_type, *args, **kwargs)


class Float(ColumnDecorator):
    column_type = types.Float


class String(ColumnDecorator):
    column_type = types.String


class Integer(ColumnDecorator):
    column_type = types.Integer


class Timestamp(ColumnDecorator):
    column_type = types.Timestamp


class Boolean(ColumnDecorator):
    column_type = types.Boolean


class Multi(ColumnDecorator):
    column_type = types.Multi


class Fulltext(object):
    sphinx_field = True

    def __init__(self, weight=None):
        self.weight = weight


class SphinxDeclarativeMeta(DeclarativeMeta):
    def __init__(cls, name, bases, dict_):
        if '_decl_class_registry' not in cls.__dict__:
            sphinx_attribute_names = set()
            sphinx_field_names = set()
            sphinx_field_weights = {}
            for base in cls.__mro__:
                for attrname, attrvalue in base.__dict__.items():
                    if hasattr(attrvalue, 'sphinx_attribute'):
                        sphinx_attribute_names.add(attrname)
                    if hasattr(attrvalue, 'sphinx_field'):
                        sphinx_field_names.add(attrname)
                        if getattr(attrvalue, 'weight', 0):
                            sphinx_field_weights[attrname] = attrvalue.weight
        super(SphinxDeclarativeMeta, cls).__init__(name, bases, dict_)


def as_sphinx_declarative(cls):
    return declarative_base(
        cls=cls,
        name=cls.__name__,
        metaclass=SphinxDeclarativeMeta,
    )



from itertools import chain

from sqlalchemy import sql
from sqlalchemy import func
from sqlalchemy.sql.expression import _literal_as_text
from sqlalchemy.sql.visitors import replacement_traverse
from sqlalchemy.orm import Query
from sqlalchemy.orm import class_mapper
from sqlalchemy.orm.base import manager_of_class
from sqlalchemy.orm.base import _generative
from sqlalchemy.orm.query import _MapperEntity

from .expression import MatchClause
from .expression import OptionsClause
from .expression import escape



class SphinxQuery(Query):

    _options = None

    def actual_count(self):
        should_nest = [self._should_nest_selectable]
        def ent_cols(ent):
            if isinstance(ent, _MapperEntity):
                return ent.mapper.primary_key
            else:
                should_nest[0] = True
                return [ent.column]

        return self._col_aggregate(
            sql.literal_column('*'),
            sql.func.count,
            nested_cols=chain(*[ent_cols(ent) for ent in self._entities]),
            should_nest=should_nest[0]
        )

    def count(self):
        self._clone().limit(0).all()
        meta = self.session.execute('SHOW META').fetchone()
        if meta and len(meta) == 2 and meta[0] == 'total':
            return int(meta[1])
        return 0

    @_generative(Query._no_statement_condition, Query._no_limit_offset)
    def options(self, **kwargs):
        if self._options is None:
            self._options = OptionsClause(kwargs)
        else:
            self._options.extend(kwargs)

    @_generative(Query._no_statement_condition, Query._no_limit_offset)
    def match(self, *args, **kwargs):
        clause_already_exists = [False]
        def replace(node):
            if isinstance(node, MatchClause):
                clause_already_exists[0] = True
                node.extend(args, kwargs)
            return node

        self._criterion = replacement_traverse(self._criterion, {}, replace)
        if not clause_already_exists[0]:
            criterion = _literal_as_text(MatchClause(args, kwargs))
            self._criterion = self._adapt_clause(criterion, True, True)

    def snippets(self, data, term, **options):
        options.setdefault('query_mode', 1)
        options.setdefault('limit_passages', 1)
        options.setdefault('weight_order', 1)
        options.setdefault('limit', 500)
        options.setdefault('around', 10)
        options.setdefault('before_match', u'<span class="highlight">')
        options.setdefault('after_match', u'</span>')
        model = self._primary_entity.type
        engine = self.session.get_bind(model)

        _param_name_n = [0]

        def gen_param_name():
            _param_name_n[0] += 1
            return u'param{}'.format(_param_name_n[0])

        if not isinstance(data, (list, tuple)):
            data = [data]

        bindparams = [
            sql.bindparam(gen_param_name(), model.__tablename__),
            sql.bindparam(gen_param_name(), escape(term)),
        ]

        options_clause_list = []
        for name, value in options.iteritems():
            param = sql.bindparam(gen_param_name(), value)
            bindparams.append(param)
            options_clause_list.append(u'{value} AS {name}'.format(
                name=name,
                value=unicode(param),
            ))

        options_clause = u', '.join(options_clause_list)

        data_clause_list = []
        for chunk in data:
            param = sql.bindparam(gen_param_name(), chunk)
            bindparams.append(param)
            data_clause_list.append(unicode(param))

        data_clause = u'({})'.format(u', '.join(data_clause_list))

        raw_sql = u'CALL SNIPPETS({data}, {index}, {term}, {options})'.format(
            data=data_clause,
            index=unicode(bindparams[0]),
            term=unicode(bindparams[1]),
            options=options_clause,
        )

        return engine.execute(sql.text(raw_sql, bindparams=bindparams))

    def _simple_statement(self, context):
        statement = super(SphinxQuery, self)._simple_statement(context)
        statement._options = self._options
        return statement

    def proxy(self, session, cls, cls_identity='id'):
        items = list(self)
        identities = [getattr(item, cls_identity) for item in items]
        weights = {getattr(item, cls_identity): item.weight for item in items}
        return BulkIdProxy(session, identities,cls, cls_identity, weights)


class BulkIdProxy(object):

    def __init__(self, session, identities, cls, key=None, weights=None):
        self._session = session
        self._cls = cls
        self._weights = weights or {}
        self._identities = identities
        self._options = []
        if isinstance(key, basestring):
            manager = manager_of_class(cls)
            self._key = manager[key]
        elif key is None:
            mapper = class_mapper(cls)
            primary_keys = mapper.primary_key
            assert len(primary_keys) == 1
            self._key = primary_keys[0]
        else:
            self._key = key

    def options(self, *args):
        self._options.extend(args)
        return self

    def count(self):
        return len(self._identities)

    def __getitem__(self, item):
        keys = self._identities[item]
        if not keys:
            return []
        query = self._session.query(self._cls)\
            .filter(self._key.in_(keys)) \
            .order_by(func.field(self._key, *keys))
        if self._options:
            query = query.options(*self._options)
        items = list(query)
        for item in items:
            item.__weight__ = self._weights.get(item.id, None)
        return items



import asyncio

from sqlalchemy.dialects.mysql.pymysql import MySQLDialect_pymysql
from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import DBAPIError

import aiomysql
import aiomysql.sa
import aiomysql.utils


_dialect = MySQLDialect_pymysql(
    paramstyle='pyformat',
    dbapi=MySQLDialect_pymysql.dbapi(),
)
_dialect.default_paramstyle = 'pyformat'


class SAConnection(aiomysql.sa.SAConnection):
    @asyncio.coroutine
    def _execute(self, query, *multiparams, **params):
        try:
            result = yield from super()._execute(query, *multiparams, **params)
            return result
        except self._dialect.dbapi.Error as e:
            if isinstance(query, str):
                statement = query
                params = params
            else:
                compiled = query.compile(dialect=self._dialect)
                statement = str(compiled)
                params = compiled.params
            raise DBAPIError.instance(
                statement,
                params,
                e,
                self._dialect.dbapi.Error,
                dialect=self._dialect,
            )


class Engine(aiomysql.sa.Engine):

    @asyncio.coroutine
    def _acquire(self):
        raw = yield from self._pool.acquire()
        return SAConnection(raw, self)



def create_engine(url, **kwargs):
    url_kwargs = make_url(url).translate_connect_args(
        database='db',
        username='user',
    )
    kwargs.update(url_kwargs, **url.query)
    return aiomysql.utils._PoolContextManager(_create_engine(**kwargs))


@asyncio.coroutine
def _create_engine(minsize=1, maxsize=10, loop=None, **kwargs):
    if loop is None:
        loop = asyncio.get_event_loop()
    pool = yield from aiomysql.create_pool(minsize=minsize, maxsize=maxsize,
                                           loop=loop, **kwargs)
    conn = yield from pool.acquire()
    try:
        return Engine(_dialect, pool, **kwargs)
    finally:
        pool.release(conn)




import asyncio

from ikcms.ws_apps.base.exceptions import ClientError

def asynctest(coroutine):
    """ Asynctest decorator """
    def wrapper(self):
        async def awrapper(self):
            asetup = getattr(self, 'asetup', None)
            kwargs = asetup and await asetup() or {}
            await coroutine(self, **kwargs)
            aclose = getattr(self, 'aclose', None)
            if aclose:
                await aclose(**kwargs)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(awrapper(self))
    wrapper.__name__ = coroutine.__name__
    wrapper.coroutine = coroutine
    return wrapper






class ServerBase:
    def serve_forever(self):
        raise NotImplementedError

    def get_remote_address(self, client_id):
        raise NotImplementedError

    async def send(self, client_id, message):
        raise NotImplementedError

    async def recv(self, client_id):
        raise NotImplementedError

    async def disconnect(self, client_id, code=1000, reason=''):
        raise NotImplementedError

    async def ping(self, client_id):
        raise NotImplementedError



import json
from unittest import TestCase
from unittest.mock import MagicMock

from ikcms.utils.asynctests import asynctest

from ikcms.ws_apps.base import App
from ikcms.ws_apps.base import exceptions
from ikcms.ws_apps.base.client import Client


class SendException(Exception):
    pass

class RecvException(Exception):
    pass


class ServerMock:

    def __init__(self, recvs=None, send_error=False, onrecv=None, onsend=None):
        self.recvs = recvs and recvs.copy() or []
        assert isinstance(self.recvs, list)
        self.sends = []
        self.send_error = send_error
        self.onrecv = onrecv
        self.onsend = onsend

    async def recv(self, client_id):
        if self.onrecv:
            onrecv()
        try:
            recv = self.recvs.pop()
        except IndexError:
            raise RecvException
        return json.dumps(recv)

    async def send(self, client_id, json_data):
        if self.onsend:
            onsend()
        if self.send_error:
            raise SendException
        self.sends.append(json.loads(json_data))


class ClientMock(dict):

    def __init__(self, app, server, client_id):
        self['app'] = app
        self['server'] = server
        self['client_id'] = client_id
        self['closed'] = False

    async def close(self):
        self['closed'] = True


class AppMockMixin:

    def __init__(self, **kwargs):
        self.log = []
        self.__dict__.update(kwargs)

    async def add_client(self, server, client_id):
        self.log.append(('add_client', (server, client_id)))
        return 'Client:' + client_id

    async def remove_client(self, client_id):
        self.log.append(('remove_client', (client_id,)))

    async def handle_connection_error(self, client, exc):
        if isinstance(exc, (SendException, RecvException)):
            self.log.append(('handle_connection_error', (client, exc.__class__)))
        else:
            raise


class AppMock(AppMockMixin, App):
    
    pass


class AppTestCase(TestCase):

    App = App
    AppMock = AppMock
    Client = Client

    def get_app(self, cfg=None, App=None):
        cfg = cfg or MagicMock()
        App = App or self.App
        return App(cfg)

    @asynctest
    async def test_init(self):
        app = self.get_app(cfg='test_cfg')
        self.assertEqual(app.cfg, 'test_cfg')
        self.assertEqual(app.clients, {})
        self.assertEqual(app.client_class, self.Client)
        self.assertEqual(app.handlers, {})

    @asynctest
    async def test_add_client(self):
        client_ids = ['client_id1', 'client_id2', 'client_id3']
        app = self.get_app()
        server = ServerMock()
        app.client_class = ClientMock

        test_clients = {}
        for client_id in client_ids:
            test_clients[client_id] = ClientMock(app, server, client_id)
            await app.add_client(server, client_id)
            self.assertEqual(app.clients, test_clients)

            with self.assertRaises(exceptions.ClientAlreadyAddedError) as ctx:
                await app.add_client(server, client_id)
            self.assertEqual(ctx.exception.kwargs['client_id'], client_id)

    @asynctest
    async def test_remove_client(self):
        app = self.get_app()
        server = ServerMock()
        client_ids = ['client_id1', 'client_id2', 'client_id3']
        test_clients = {}
        for client_id in client_ids:
            test_clients[client_id] = ClientMock(app, server, client_id)
 
        app = self.get_app()
        app.clients = test_clients.copy()
        for client_id in client_ids:
            client = test_clients.pop(client_id)
            await app.remove_client(client_id)
            self.assertEqual(app.clients, test_clients)


            with self.assertRaises(exceptions.ClientNotFoundError) as ctx:
                await app.remove_client('error_client')
            self.assertEqual(ctx.exception.kwargs['client_id'], 'error_client')

            with self.assertRaises(exceptions.ClientNotFoundError) as ctx:
                await app.remove_client(client_id)
            self.assertEqual(ctx.exception.kwargs['client_id'], client_id)


    @asynctest
    async def test_call(self):
        #handle
        test_request = {
            'name': 'request',
            'request_id': 'test_id',
            'handler': 'test_handler',
            'body': {},
        }
        test_response = {
            'name': 'response',
            'request_id': 'test_id',
            'handler': 'test_handler',
            'body': {},
        }
        async def handle(client, request):
            self.assertEqual(client, 'Client:client1')
            self.assertEqual(request, test_request)
            return test_response
        app_mock = self.AppMock(handle=handle)
        server = ServerMock([test_request])
        await App.__call__(app_mock, server, 'client1')
        self.assertEqual(server.sends, [test_response])
        self.assert_connection_error(app_mock, server, RecvException)

        # recv error
        async def handle(client, request):
            self.assertTrue(False)
        app_mock = self.AppMock(handle=handle)
        server = ServerMock()
        await App.__call__(app_mock, server, 'client1')
        self.assert_connection_error(app_mock, server, RecvException)


        #send error
        test_request = {'test_request2':567}
        test_response = {'test_response2': '222'}
        async def handle(client, request):
            self.assertEqual(client, 'Client:client1')
            self.assertEqual(request, test_request_dict)
            return test_response_dict
        app_mock = self.AppMock(handle=handle)
        server = ServerMock([test_request], send_error=True)
        await App.__call__(app_mock, server, 'client1')
        self.assert_connection_error(app_mock, server, SendException)

        #server error
        test_request = {
            'name': 'request',
            'handler': 'test_handler',
            'request_id': 'test_request_id',
            'body': {},
        }
        test_response = {
            'name': 'error',
            'request_id': 'test_request_id',
            'handler': 'test_handler',
            'body': {
                'error': 'InternalServerError',
                'message': 'Internal Server Error',
                'kwargs': {},
            },
        }
        async def handle(client, request):
            self.assertEqual(client, 'Client:client1')
            self.assertEqual(request, test_request)
            raise exceptions.BaseError(test='rrr555')
        app_mock = self.AppMock(handle=handle)
        server = ServerMock([test_request])
        await App.__call__(app_mock, server, 'client1')
        self.assertEqual(server.sends[0], test_response)
        self.assert_connection_error(app_mock, server, RecvException)

        #client error
        test_request = {
            'name': 'request',
            'handler': 'test_handler',
            'request_id': 'test_request_id',
            'body': {},
        }
        test_response = {
            'name': 'error',
            'request_id': 'test_request_id',
            'handler': 'test_handler',
            'body': {
                'error': 'BaseError',
                'message': 'Error message',
                'kwargs': {'test': 'rrr555'},
            },
        }
        async def handle(client, request):
            self.assertEqual(client, 'Client:client1')
            self.assertEqual(request, test_request)
            raise exceptions.ClientError(exceptions.BaseError(test='rrr555'))
        app_mock = self.AppMock(handle=handle)
        server = ServerMock([test_request])
        await App.__call__(app_mock, server, 'client1')
        self.assertEqual(server.sends[0], test_response)
        self.assert_connection_error(app_mock, server, RecvException)

        #Handle RequestTypeError
        app_mock = self.AppMock(handle=handle)
        requests = [[], "", 17, None]
        server = ServerMock(requests, [])
        await App.__call__(app_mock, server, 'client1')
        self.assert_connection_error(app_mock, server, RecvException)
        for s in server.sends:
            self.assertEqual(s['name'], 'error')
            self.assertEqual(s['handler'], '')
            self.assertEqual(s['request_id'], '')
            self.assertEqual(s['body']['error'], 'RequestTypeError')


        #Handle MessageError field 'name'
        app_mock = self.AppMock(handle=handle)
        test_request = {'body': {}}
        error_names = [None, 'erw;df', 45, 5.6, {'d':2}, [3, 5]]
        requests = [dict(test_request, name=name) for name in error_names]
        requests.append(test_request.copy())
        server = ServerMock(requests, [])
        await App.__call__(app_mock, server, 'client1')
        self.assert_connection_error(app_mock, server, RecvException)
        for s in server.sends:
            self.assertEqual(s['name'], 'error')
            self.assertEqual(s['handler'], '')
            self.assertEqual(s['request_id'], '')
            self.assertEqual(s['body']['error'], 'MessageFieldsError')
            self.assertEqual(
                list(s['body']['kwargs']['errors']), ['name'],
            )

        #Handle MessageError field 'body'
        app_mock = self.AppMock(handle=handle)
        test_request = {'name': 'ok'}
        error_names = [None, 'erw;df', 45, 5.6, [3, 5]]
        requests = [dict(test_request, body=body) for body in error_names]
        requests.append(test_request.copy())
        server = ServerMock(requests, [])
        await App.__call__(app_mock, server, 'client1')
        self.assert_connection_error(app_mock, server, RecvException)
        for s in server.sends:
            self.assertEqual(s['name'], 'error')
            self.assertEqual(s['handler'], '')
            self.assertEqual(s['request_id'], '')
            self.assertEqual(s['body']['error'], 'MessageFieldsError')
            self.assertEqual(
                list(s['body']['kwargs']['errors']), ['body'],
            )

        #Handle MessageError field 'request_id'
        app_mock = self.AppMock(handle=handle)
        test_request = {'name': 'request', 'handler': 'ok', 'body':{}}
        error_request_id = [None, 45, 5.6, {'d':2}, [3, 5]]
        requests = [dict(test_request, request_id=r_id) \
            for r_id in error_request_id]
        requests.append(test_request.copy())
        server = ServerMock(requests, [])
        await App.__call__(app_mock, server, 'client1')
        self.assert_connection_error(app_mock, server, RecvException)
        for s in server.sends:
            self.assertEqual(s['name'], 'error')
            self.assertEqual(s['handler'], '')
            self.assertEqual(s['request_id'], '')
            self.assertEqual(s['body']['error'], 'MessageFieldsError')
            self.assertEqual(
                list(s['body']['kwargs']['errors']), ['request_id'],
            )


        #Handle MessageError field 'handler'
        app_mock = self.AppMock(handle=handle)
        test_request = {'name': 'request', 'request_id': 'ok', 'body':{}}
        error_handlers = [None, 45, 5.6, {'d':2}, [3, 5]]
        requests = [dict(test_request, handler=h) for h in error_handlers]
        requests.append(test_request.copy())
        server = ServerMock(requests, [])
        await App.__call__(app_mock, server, 'client1')
        self.assert_connection_error(app_mock, server, RecvException)
        for s in server.sends:
            self.assertEqual(s['name'], 'error')
            self.assertEqual(s['handler'], '')
            self.assertEqual(s['request_id'], '')
            self.assertEqual(s['body']['error'], 'MessageFieldsError')
            self.assertEqual(
                list(s['body']['kwargs']['errors']), ['handler'],
            )

    @asynctest
    async def test_handle(self):
        app = self.get_app()
        async def test_handler(client, request):
            return {'test':'ok', 'request': request}
        app.handlers = {'test_handler': test_handler}
        request = {
            'name': 'request',
            'request_id': 'test_request_id',
            'handler': 'test_handler',
            'body': {'test': 12345},
        }
        result = await app.handle('client5', request.copy())
        self.assertEqual(
            result,
            {
                'name': 'response',
                'request_id': 'test_request_id',
                'handler': 'test_handler',
                'body': {'test': 'ok', 'request': {'test': 12345}},
            },
        )

        #Handler not allowed error
        request = {
            'name': 'request',
            'handler': 'error_handler',
            'request_id':'xxx',
            'body': {},
        }
        with self.assertRaises(exceptions.ClientError) as ctx:
            result = await app.handle('client5', request)
        exc = ctx.exception
        self.assertEqual(exc.error, 'HandlerNotAllowedError')
        self.assertEqual(exc.kwargs, {'handler': 'error_handler'})

    def assert_connection_error(self, app_mock, server, exc_class):
        self.assertEqual(
            app_mock.log,
            [
                ('add_client', (server, 'client1')),
                ('handle_connection_error', ('Client:client1', exc_class)),
                ('remove_client', ('client1',)),
            ]
        )





from . import exceptions
from .base import Request
from .base import Response
from .handlers import WebHandler
from .handlers import h_not_found
from .handlers import h_server_error
from .handlers import h_match
from .handlers import h_prefix
from .handlers import h_cases
from .handlers import h_static_files
from .handlers import h_domain
from .handlers import h_subdomain
from .handlers import h_app
from .handlers import h_namespace
from .view import BaseView
from .view import HView
from .view import viewhandler
from .view import viewfilter



class Component(object):

    name = None

    def __init__(self, app):
        assert self.name, \
            'Component {} has not name property'.format(self)
        assert not hasattr(app, self.name), \
            'Component with name "{}" already exists'.format(self.name)

        self.app = app
        setattr(app, self.name, self)

    @classmethod
    def create(cls, app):
        return cls(app)

    def on_initialization_end(self):
        pass

    def on_request(self, request):
        pass

    def on_init_env(self, env):
        pass

    def on_close_env(self, env):
        pass

    @classmethod
    def create_cls(cls, **kwargs):
        return type(cls.__name__, (cls,), kwargs)



import ikcms.apps.base
import ikcms.apps.composite


class App(ikcms.apps.base.App):

    components = []

    def __init__(self, cfg):
        self.cfg = cfg
        self.cfg.config_uid()
        self.cfg.config_logging()
        self.env_class = self.get_env_class()
        self.components = [
            component.create(self) for component in self.components]
        self.handler = self.get_handler()
        self.root = self.get_root()
        for component in self.components:
            component.on_initialization_end()

    def get_request(self, environ):
        request = self.Request(environ, charset='utf-8')
        for component in self.components:
            component.on_request(request)
        return request

    def get_env_class(self):
        from .env import Environment
        return Environment

    def get_handler(self):
        raise NotImplementedError




# coding: utf8
from iktomi.cms.stream_actions import GetAction
from ikcms.utils import cached_property
from ikcms.utils import N_
from ikcms.web import h_prefix


class CfgMixin(object):

    PREVIEW = True
    DOMAINS_AS_PATH = True

    def __init__(self, **kwargs):
        assert 'parent_app' in kwargs, 'parent_app property required'
        super(CfgMixin, self).__init__(**kwargs)

    @cached_property
    def ROOT_DIR(self):
        return self.parent_app.cfg.ROOT_DIR

    @cached_property
    def DATABASES(self):
        return self.parent_app.cfg.DATABASES

    @cached_property
    def DATABASE_PARAMS(self):
        return self.parent_app.cfg.DATABASE_PARAMS

    @cached_property
    def STATIC_ENABLED(self):
        return self.parent_app.cfg.STATIC_ENABLED

    @cached_property
    def STATIC_URL(self):
        return self.parent_app.cfg.PREVIEW_STATIC_URL

    @cached_property
    def MEDIA_DIR(self):
        return self.parent_app.cfg.MEDIA_ROOT_ADMIN

    @cached_property
    def MEDIA_URL(self):
        return self.parent_app.cfg.MEDIA_URL

    @cached_property
    def TMP_DIR(self):
        return self.parent_app.cfg.TMP_DIR

    @cached_property
    def DEFAULT_CUSTOM_CFG(self):
        return self.parent_app.cfg.DEFAULT_CUSTOM_CFG

    @cached_property
    def REDIS_URL(self):
        return self.parent_app.cfg.REDIS_URL


class AppMixin(object):

    PREVIEW = True

    def get_handler(self):
        return h_prefix('/preview') | super(AppMixin, self).get_handler()

    def preview_buttons(self, env, stream_id, item_id, **kwargs):
        admin_env = env.parent_env
        stream = admin_env.streams[stream_id]
        if not stream.get_permissions(admin_env):
            return {}

        result = {}
        result['data-preview-title'] = kwargs.get('title', u'Редактировать')
        result['data-preview-where'] = kwargs.get('where', 'bottom')
        result['data-preview-position'] = kwargs.get('position', 'relative')
        result['data-preview-left'] = kwargs.get('left')
        result['data-preview-right'] = kwargs.get('right')
        result['data-preview-top'] = kwargs.get('top')
        result['data-preview-bottom'] = kwargs.get('bottom')
        if kwargs.get('hidden', False):
            result['data-preview-hidden'] = 1
        url_kwargs = {'item': item_id}
        if 'lang' in kwargs:
            url_kwargs['lang'] = kwargs['lang']
        item_url = stream.url_for(admin_env, 'item', **url_kwargs)
        result['data-preview-edit'] = item_url
        return result


class PreviewStreamAction(GetAction):

    item_lock = True
    allowed_for_new = False
    cls = 'preview'
    action = 'preview'
    title = u'Превью'
    mode = 'internal'
    get_url = lambda env, item: env.url_for_obj(item)

    def __init__(self, stream=None, **kw):
        super(PreviewStreamAction, self).__init__(stream, **kw)
        self.get_url = kw.get('get_url', self.get_url)

    @property
    def app(self):
        return self

    def url(self, env, item):
        preview_app = env.app.preview.app
        preview_env = preview_app.get_env(env.request)
        if hasattr(env, 'lang'):
            preview_app.i18n.set_lang(preview_env, env.lang)
        try:
            return self.get_url(preview_env, item)
        finally:
            preview_env.close()

    def external_url(self, env, data, item):
        return '#'

    def is_available(self, env, item):
        return GetAction.is_available(self, env, item) and \
                getattr(env, 'version', None) != 'front' and \
                self.url(env, item)

    def preview(self, env, data):
        return None

    __call__ = preview





import aioredis
from . import base


class Component(base.Component):

    DEFAULT_REDIS_HOST = 'localhost'
    DEFAULT_REDIS_PORT = 6379

    def __init__(self, app, redis):
        super().__init__(app)
        self.redis = redis

    @classmethod
    async def create(cls, app):
        host = getattr(app.cfg, 'REDIS_HOST', cls.DEFAULT_REDIS_HOST)
        port = getattr(app.cfg, 'REDIS_PORT', cls.DEFAULT_REDIS_PORT)
        address = (host, port)
        redis = await aioredis.create_redis(address)
        return cls(app, redis)

    async def get(self, key):
        return await self.redis.get(key)

    async def set(self, key, value, expire=0):
        return await self.redis.set(key, value, expire=expire)

    async def delete(self, key):
        return await self.redis.delete(key)


component = Component.create_cls



import lxml
import jinja2.ext
from jinja2 import Markup

from . import tags


__all__ = (
    'Tag',
    'DropEmptyTags'
    'media',
    'photo',
    'photoset',
    'video',
    'drop_empty_p',
)



def replace_xml_tag_with_text(xml_tag, value):
    parent = xml_tag.getparent()
    if parent is not None:
        fragments = lxml.html.fragments_fromstring(value)
        if fragments and isinstance(fragments[0], basestring):
            text = fragments.pop(0)
            index = parent.index(xml_tag)
            if index > 0:
                element = parent.getchildren()[index - 1]
                element.tail = (element.tail or '') + text
            else:
                parent.text = (parent.text or '') + text
        for fragment in fragments:
            parent.insert(parent.index(xml_tag), fragment)
        xml_tag.drop_tree()


class Tag(object):

    def __init__(self, selector, replacement_tag):
        assert issubclass(replacement_tag, jinja2.ext.Extension), \
            'jinja2.ext.Extension subclass trequired'
        self.selector = selector
        self.replacement_tag = replacement_tag

    def __call__(self, component, root, **kwargs):
        selector = self.selector.format(**kwargs)
        for xml_tag in root.xpath(selector):
            self.replace(component, xml_tag, **kwargs)

    def replace(self, component, xml_tag, **kwargs):
        #replacement_tag = self.replacement_tag.format(**kwargs)
        jinja_env = component.app.render._env
        tag = jinja_env.extensions.get(self.replacement_tag.identifier)
        if not tag:
            jinja_env.add_extension(self.replacement_tag)
            tag = jinja_env.extensions[self.replacement_tag.identifier]
        tag_attrs = dict(xml_tag.attrib, **kwargs)
        def caller():
            return Markup((xml_tag.text or '') + ''.join(
                [lxml.html.tostring(child) for child in xml_tag.iterchildren()],
            ))
        tag_attrs['caller'] = caller
        replace_xml_tag_with_text(xml_tag, tag.render(**tag_attrs))


class DropEmptyTags(object):

    def __init__(self, *selectors):
        self.selectors = selectors

    def __call__(self, component, root, **kwargs):
        for selector in self.selectors:
            for xml_tag in root.xpath(selector):
                if self._is_empty(xml_tag):
                    self._drop(xml_tag)

    def _is_empty(node):
        return not node.text_content().strip()

    def _drop(node):
        return node.getparent().remove(node)


media = Tag('//iktomi_media', tags.iktomi_media)
photo = Tag('//iktomi_photo', tags.iktomi_photo)
photoset = Tag('//iktomi_photoset', tags.iktomi_photoset)
video = Tag('//iktomi_video', tags.iktomi_video)
drop_empty_p = DropEmptyTags('//p')





__all__ = (
    'ValidationError',
    'RawValueTypeError',
)

class BaseError(Exception):

    message = 'Error message'
    kwargs = {}

    def __init__(self, **kwargs):
        self.kwargs = kwargs.copy()

    @property
    def error(self):
        return self.__class__.__name__

    def __str__(self):
        return self.message.format(**self.kwargs)


class ValidationError(BaseError):

    message = '{error}'

    def __init__(self, error):
        super().__init__(error=error)


class BaseValueError(BaseError):

    message = 'Field "{field_name}" error: {error}'

    def __init__(self, field_name, error):
        super().__init__(field_name=field_name, error=error)

    def add_name(self, name):
        field_name = self.kwargs.get(field_name)
        if field_name:
            self.kwargs['field_name'] = '{}.{}'.format(name, field_name)
        else:
            self.kwargs['field_name'] = name


class RawValueError(BaseValueError):

    pass


class RawValueRequiredError(RawValueError):

    def __init__(self, field_name):
        super().__init__(field_name, 'Field required')


class RawValueNoneNotAllowedError(RawValueError):

    def __init__(self, field_name):
        super().__init__(field_name, 'None value not allowed')


class RawValueTypeError(RawValueError):

    def __init__(self, field_name, field_type):
        super().__init__(
            field_name,
            'Required type="{}"'.format(field_type),
        )
        self.kwargs['field_type'] = field_type


class PythonValueError(BaseValueError):
    pass


class PythonValueRequiredError(PythonValueError):

    def __init__(self, field_name):
        super().__init__(
            field_name,
            'Field required'
        )


class PythonValueNoneNotAllowedError(PythonValueError):

    def __init__(self, field_name):
        super().__init__(
            field_name,
            'None value not allowed'
        )


class PythonValueTypeError(PythonValueError):

    def __init__(self, field_name, field_type):
        super().__init__(
            field_name,
            'Required type="{}"'.format(field_type),
        )
        self.kwargs['field_type'] = field_type




from datetime import datetime

from sqlalchemy import Column
from sqlalchemy import String
from sqlalchemy import DateTime
from iktomi.db.sqla.types import StringList
from . import encrypt_password

from ikcms.orm import mappers
from ikcms.orm import relations

__all__ = (
    'AdminGroup',
    'AdminUser',
)

class AdminGroup(mappers.Base):

    name = 'AdminGroup'

    def create_columns(self):
        return [
            Column('title', String(200), nullable=False),
            Column('roles', StringList(1000), nullable=False, default=''),
        ]

    def create_id_column(self):
        return Column('id', String(20), primary_key=True, nullable=False)

    async def schema_initialize(self, session):
        print('Adding Administrators group')
        cnt = await self.query().id('admins').count_items(session)
        if not cnt:
            admin_group = dict(
                id='admins',
                title='Administrators',
            )
            await self.query().insert_item(session, admin_group)


class AdminUser(mappers.Base):

    name = 'AdminUser'

    def create_columns(self):
        return [
            Column('email', String(200)),
            Column('name', String(200), nullable=False),
            Column('login', String(200), nullable=False, unique=True),
            Column('password', String(200), nullable=False),
            Column('created_dt', DateTime, nullable=False, default=datetime.now),
        ]

    def create_relations(self):
        return {'groups': relations.M2M(self, 'AdminGroup')}

    async def schema_initialize(self, session):
        print('Adding root user')
        cnt = await self.query().filter_by(login='root').count_items(session)
        if not cnt:
            root_user = dict(
                login='root',
                password=self.encrypt_password('root'),
                name='Administrator',
                groups=['admins'],
            )
            await self.query().insert_item(session, root_user)

    def encrypt_password(self, password):
        return encrypt_password(password)




from unittest import TestCase, skipIf
from unittest.mock import MagicMock

from tests.cfg import cfg

if cfg.AIO_DB_ENABLED:
    from ikcms.ws_components.db import component
    from ikcms.orm import mappers
    from tests.models import create_models1, create_models2, create_metadata

from ikcms.utils.asynctests import asynctest
from iktomi.utils import cached_property



@skipIf(not cfg.AIO_DB_ENABLED, 'AIO DB DISABLED')
class SQLAComponentTestCase(TestCase):

    @cached_property
    def test_models(self):
        return {
            'db1': create_models1(),
            'db2': create_models2(),
        }

    @skipIf(not cfg.AIOMYSQL_ENABLED, 'Aiomysql not installed')
    @asynctest
    async def test_mysql(self):
        await self._db_test(cfg.MYSQL_URL)

    @skipIf(not cfg.AIOPG_ENABLED, 'Aiopg not instaled')
    @asynctest
    async def test_postgress(self):
        await self._db_test(cfg.POSTGRESS_URL)

    async def _db_test(self, db_url):
        app = MagicMock()
        del app.db
        app.cfg.DATABASES = {
            'db1': db_url,
            'db2': db_url,
        }

        session_cls_mock = MagicMock()
        Component = component(
            mappers=mappers.Registry(create_metadata(self.test_models.values())),
            session_cls=session_cls_mock,
        )
        db = await Component.create(app)
        self.assertEqual(set(db.engines.keys()), set(app.cfg.DATABASES.keys()))
        self.assertEqual(
            db.binds,
            {
                self.test_models['db1'].test_table1: db.engines['db1'],
                self.test_models['db2'].test_table2: db.engines['db2'],
            }
        )
        session = await db()
        self.assertEqual(session, session_cls_mock(db.engines, db.binds))



from ikcms.forms import fields


__all__ = (
    'id',
    'title',
)


class id(fields.Int):
    name = 'id'
    label = 'Id'


class title(fields.String):

    name = 'title'
    label = 'Title'


class Date(fields.Date):
    pass



import logging
import cPickle as pickle
import time

import sqlalchemy as sa
from sqlalchemy.orm import Query

import ikcms.components.base


logger = logging.getLogger(__name__)


class Component(ikcms.components.base.Component):

    name = 'cached_tree'
    model = 'front.TreeItem'
    check_timeout = 3
    lock_timeout = 3

    def __init__(self, app):
        super(Component, self).__init__(app)
        self.model = self.app.db.get_model(self.model)
        self.cache_key_checked_ts = '{}:checked_ts'.format(self.name)
        self.cache_key_updated_ts = '{}:updated_ts'.format(self.name)
        self.cache_key_updating = '{}:updating'.format(self.name)
        self.cache_key_meta = '{}:meta'.format(self.name)
        self.cache_key_body = '{}:body'.format(self.name)
        self.cache_key_lock = '{}:lock'.format(self.name)
        self.init_cache()

    def on_request(self, request):
        self.update_cache()

    def reset_cache(self, pipe):
        self.app.cache.delete(
            self.cache_key_updated_ts,
            self.cache_key_updating,
            self.cache_key_meta,
            self.cache_key_body,
        )

    def init_cache(self):
        with self.app.cache.lock(
            self.cache_key_lock,
            expires=self.lock_timeout,
        ) as lock:
            self.update_cache()

    def update_cache(self):
        # check update timeout
        cache_checked_ts, cache_updated_ts = self.app.cache.mget(
            self.cache_key_checked_ts,
            self.cache_key_updated_ts,
        )
        try:
            cache_checked_ts = int(cache_checked_ts)
            cache_updated_ts = int(cache_updated_ts)
        except (TypeError, ValueError):
            cache_checked_ts = None
            cache_updated_ts = None

        # if check timeout not expired, updating is not required
        if cache_checked_ts:
            if time.time() < (cache_checked_ts + self.check_timeout):
                return cache_updated_ts

        # if other process is already updating cache, we do nothing
        if not self.app.cache.add(self.cache_key_updating, 1, self.lock_timeout):
            return cache_updated_ts

        now_ts = int(time.time())
        try:
            db_updated_ts = self.get_updated_ts_from_db()
        except sa.exc.ProgrammingError as exc:
            logger.warning('Retrieve {} error: {}'.format(self.model, exc))
            return None

        # if db not changed, we update cache checked ts
        if cache_updated_ts and db_updated_ts <= cache_updated_ts:
            with self.app.cache.pipe() as pipe:
                pipe.set(self.cache_key_checked_ts, now_ts)
                pipe.delete(self.cache_key_updating)
                pipe.execute()
            return cache_updated_ts

        # Get sections from db
        items_meta, items_body = self.get_items_from_db()

        items_meta = {s_id: self._dumps(s) \
            for s_id, s in items_meta.items()}
        items_body = {s_id: self._dumps(s) \
            for s_id, s in items_body.items()}

        # Update cache 
        with self.app.cache.pipe() as pipe:
            pipe.set(self.cache_key_updated_ts, db_updated_ts)
            pipe.set(self.cache_key_checked_ts, now_ts)
            pipe.delete(self.cache_key_updating)
            pipe.delete(self.cache_key_meta)
            pipe.delete(self.cache_key_body)
            pipe.hmset(self.cache_key_meta, items_meta)
            if items_body:
                pipe.hmset(self.cache_key_body, items_body)
            pipe.execute()
        return db_updated_ts

    def get_updated_ts_from_db(self):
        session = self.app.db()
        s = sa.sql.select([sa.func.max(self.model.updated_dt)])
        result = list(session.execute(s))
        session.close()
        if result and result[0][0]:
            return int(time.mktime(result[0][0].timetuple()))
        else:
            return None

    def get_items_from_db(self):
        session = self.app.db()
        db_objs = self._get_objs_from_db(session)

        objs_by_id = {obj.id: obj for obj in db_objs}
        items = [obj.to_meta_dict() for obj in db_objs]
        items_by_id = {item['id']: item for item in items}
        items_by_parent = {}
        for item in items:
            if item['parent_id'] is not None:
                parent_item = items_by_id.get(item['parent_id'])
                if not parent_item:
                    continue
            items_by_parent.setdefault(item['parent_id'], []).append(item)

        items = self._walk_tree(items_by_parent)
        items = self._filter_items(items)
        items_meta = {item['id']: item for item in items}
        items_body = {item_id: objs_by_id[item_id].to_body_dict()\
            for item_id in items_meta}
        root_ids = [item['id'] for item in items_by_parent.get(None, [])]
        items_meta[''] = {'children': root_ids}
        session.close()
        return items_meta, items_body

    def get_items(self, ids):
        return self._get_items_meta(ids)

    def get_item(self, id):
        return self.get_items([id])[0]

    def get_items_with_body(self, db, ids):
        with self.app.cache.pipe() as pipe:
            metas = self._get_items_meta(ids)
            bodies = self._get_items_bodies(ids)
        items = []
        for meta, body in zip(metas, bodies):
            if meta is not None and body is not None:
                item = dict(meta, **body)
            else:
                item = None
            items.append(item)
        return items

    def get_item_with_body(self, db, id):
        return self.get_items_with_body(db, [id])[0]

    def _get_items_meta(self, ids):
        if not ids:
            return []
        raw_items = self.app.cache.hmget(self.cache_key_meta, ids)
        items = []
        for id, item in zip(ids, raw_items):
            if item is not None:
                item = self._loads(item)
            items.append(item)
        return items

    def _get_items_bodies(self, ids):
        if not ids:
            return []
        raw_items = self.app.cache.hmget(self.cache_key_body, ids)
        items = []
        for id, item in zip(ids, raw_items):
            if item is not None:
                item = self._loads(item)
            items.append(item)
        return items

    def _walk_tree(self, items_by_parent, parent_item=None):
        result = []
        parent_id = parent_item and parent_item['id'] or None
        for item in items_by_parent.get(parent_id, []):
            item = self._create_item(item, parent_item)
            result.append(item)
            result += self._walk_tree(items_by_parent, item)
        return result

    def _create_item(self, item, parent_item):
        item = item.copy()
        item['children'] = []
        if parent_item is None:
            item['parents'] = []
        else:
            item['parents'] = parent_item['parents'] + [parent_item['id']]
            parent_item['children'].append(item['id'])
        return item

    def _filter_items(self, items):
        return items

    def _get_objs_from_db(self, session):
        return session.query(self.model).order_by(self.model.order.asc()).all()

    def _dumps(self, obj):
        return pickle.dumps(obj)

    def _loads(self, string):
        return pickle.loads(string)


component = Component.create_cls



import re

from . import exceptions


__all__ = (
    'Validator',
    'Required',
    'Regex',
    'Len',
    'Range',
)



class Validator:

    def __init__(self, field):
        self.field = field

    def __call__(self, value):
        raise NotImplementedError

    def error(self, message_name):
        message = getattr(self.field, message_name, getattr(self, message_name))
        raise exceptions.ValidationError(message.format(validator=self))


class Required(Validator):
    required_error = 'required field'

    def __init__(self, field):
        assert hasattr(field, 'required')
        super().__init__(field)
        self.required = field.required

    def __call__(self, value):
        if self.required and value in ('', [], {}, None):
            self.error('required_error')
        else:
            return value


class Regex(Validator):
    regex_error = 'field should match {validator.regex}'

    def __init__(self, field):
        assert hasattr(field, 'regex')
        super().__init__(field)
        self.regex_str = field.regex
        self.regex = field.regex and re.compile(field.regex) or None

    def __call__(self, value):
        if value is None:
            return value
        if self.regex and not self.regex.match(value):
            self.error('regex_error')
        return value


class Len(Validator):

    min_len_error = 'min length is {validator.min_len}'
    max_len_error = 'max length is {validator.max_len}'

    def __init__(self, field):
        super().__init__(field)
        assert hasattr(field, 'min_len')
        assert hasattr(field, 'max_len')
        self.min_len = field.min_len
        self.max_len = field.max_len

    def __call__(self, value):
        if value is None:
            return value
        if self.min_len is not None and len(value) < self.min_len:
            self.error('min_len_error')
        if self.max_len is not None and len(value) > self.max_len:
            self.error('max_len_error')
        return value


class Range(Validator):

    min_value_error = 'min value is {validator.min_value}'
    max_value_error = 'max value is {validator.max_value}'

    def __init__(self, field):
        super().__init__(field)
        assert hasattr(field, 'min_value')
        assert hasattr(field, 'max_value')
        self.min_value = field.min_value
        self.max_value = field.max_value

    def __call__(self, value):
        if value is None:
            return None
        if self.min_value is not None and value < self.min_value:
            self.error('min_value_error')
        if self.max_value is not None and value > self.max_value:
            self.error('max_value_error')
        return value




from unittest import TestCase
from unittest.mock import MagicMock

from ikcms.utils.asynctests import asynctest
from ikcms.ws_components.streams import streams
from ikcms.ws_components.streams import exceptions


class BaseStreamTestCase(TestCase):

    class Action(MagicMock):
        async def handle(self, env, massage):
            return '{}_handle'.format(self.name)

    class Stream(streams.Base):
        name = 'test_stream'
    Stream.actions = [Action(name='action1'), Action(name='action2')]

    stream_ids = ['test_stream']

    async def asetup(self):
        component = MagicMock(streams={})
        self.Stream.create(component, component.streams)
        return {
            'component': component,
        }

    def _get_streams(self, component):
        for id in self.stream_ids:
            self.assertTrue(id in component.streams)
            stream = component.streams[id]
            self.assertEqual(stream.id, id)
            yield stream

    async def aclose(self, component):
        pass


    @asynctest
    async def test_create(self, component):
        self._test_create(component)

    def _test_create(self, component):
        for stream in self._get_streams(component):
            self.assertIsInstance(stream, self.Stream)
            self.assertEqual(stream.component, component)
            self.assertEqual(
                stream.actions,
                [self.Stream.actions[0](stream), self.Stream.actions[1](stream)],
            )

    @asynctest
    async def test_get_action(self, component):
        for stream in self._get_streams(component):
            for action in stream.actions:
                self.assertEqual(
                    stream.get_action(action.name),
                    action,
                )
            self.assertEqual(stream.get_action('notexists'), None)
            stream.actions = []
            self.assertEqual(stream.get_action('notexists'), None)

    @asynctest
    async def test_h_action(self, component):
        for stream in self._get_streams(component):
            for action in stream.actions:
                env = MagicMock()
                message = {'action': action.name}
                result = await stream.h_action(env, message)
                self.assertEqual(result, '{}_handle'.format(action.name))

            with self.assertRaises(exceptions.StreamActionNotFoundError):
                stream.h_action(MagicMock(), {'action': 'notexists'})


class StreamTestCase(BaseStreamTestCase):

    class Stream(streams.Stream):
        name = 'test_stream'
        mapper_name = 'test_mapper'
        db_id = 'db1'

        actions = [
            BaseStreamTestCase.Action(name='action1'),
            BaseStreamTestCase.Action(name='action2'),
        ]

    test_mapper = MagicMock()
    mappers = {'db1': {'test_mapper': test_mapper}}

    async def asetup(self):
        component = MagicMock(
                app=MagicMock(
                    db=MagicMock(mappers=self.mappers),
                    auth=MagicMock(
                        get_user_perms=lambda user, permissons: user.test_perms,
                    ),
                ),
                streams={},
        )
        self.Stream.create(component, component.streams)
        return {
            'component': component,
        }

    @asynctest
    async def test_mapper_prop(self, component):
        for stream in self._get_streams(component):
            self.assertEqual(stream.mapper, self.test_mapper)
            self.assertEqual(stream.query(), self.test_mapper.query())

    @asynctest
    async def test_get_form(self, component):
        for stream in self._get_streams(component):
            stream.list_fields = [MagicMock(order=False), MagicMock(order=True)]
            form = stream.get_list_form(MagicMock())
            self.assertIsInstance(form, self.Stream.ListForm)
            self.assertEqual(form.fields, stream.list_fields)

            form = stream.get_order_form(MagicMock())
            self.assertIsInstance(form, self.Stream.ListForm)
            self.assertEqual(form.fields, [stream.list_fields[1]])


            stream.filter_fields = MagicMock()
            form = stream.get_filter_form(MagicMock())
            self.assertIsInstance(form, self.Stream.FilterForm)
            self.assertEqual(form.fields, stream.filter_fields)


            stream.item_fields = MagicMock()
            form = stream.get_item_form(MagicMock())
            self.assertIsInstance(form, self.Stream.ItemForm)
            self.assertEqual(form.fields, stream.item_fields)


class I18nStreamTestCase(StreamTestCase):
    class Stream(streams.I18nStream):
        name = 'test_stream'
        mapper_name = 'test_mapper'
        db_id = 'db1'

        actions = [
            BaseStreamTestCase.Action(name='action1'),
            BaseStreamTestCase.Action(name='action2'),
        ]

    stream_ids = ['ru.test_stream', 'en.test_stream']

    test_mappers = {'ru': MagicMock(), 'en': MagicMock()}
    mappers = {'db1': {
        'ru': {'test_mapper': test_mappers['ru']},
        'en': {'test_mapper': test_mappers['en']},
    }}

    def _test_create(self, component):
        super()._test_create(component)
        streams = list(self._get_streams(component))
        for num, stream in enumerate(streams):
            self.assertEqual(stream.lang, self.Stream.langs[num])
            self.assertEqual(
                stream.i18n_streams,
                {strm.lang: strm for strm in streams},
            )

    @asynctest
    async def test_mapper_prop(self, component):
        for stream in self._get_streams(component):
            test_mapper = self.test_mappers[stream.lang]
            self.assertEqual(stream.mapper, test_mapper)
            self.assertEqual(stream.query(), test_mapper.query())


class PubStreamTestCase(StreamTestCase):
    class Stream(streams.PubStream):
        name = 'test_stream'
        mapper_name = 'test_mapper'

        actions = [
            BaseStreamTestCase.Action(name='action1'),
            BaseStreamTestCase.Action(name='action2'),
        ]
        permissions = {
            'role1': 'rwxcd',
            'role2': 'rw',
        }

    stream_ids = ['admin.test_stream', 'front.test_stream']

    mappers = {
        'admin': {'test_mapper': MagicMock()},
        'front': {'test_mapper': MagicMock()},
    }

    @asynctest
    async def test_mapper_prop(self, component):
        for stream in self._get_streams(component):
            test_mapper = self.mappers[stream.db_id]['test_mapper']
            self.assertEqual(stream.mapper, test_mapper)
            self.assertEqual(stream.query(), test_mapper.query())

    def _test_create(self, component):
        super()._test_create(component)
        streams = list(self._get_streams(component))
        for num, stream in enumerate(streams):
            self.assertEqual(stream.db_id, self.Stream.db_ids[num])
            self.assertEqual(
                stream.pub_streams,
                {strm.db_id: strm for strm in streams},
            )
            if not num:
                self.assertEqual(
                    stream.permissions,
                    {
                        'role1': set('rwxcd'),
                        'role2': set('rw'),
                    }
                )
            else:
                self.assertEqual(
                    stream.permissions,
                    {
                        'role1': set('rx'),
                        'role2': set('r'),
                    }
                )

class PubI18nStreamTestCase(StreamTestCase):
    class Stream(streams.PubI18nStream):
        name = 'test_stream'
        mapper_name = 'test_mapper'

        actions = [
            BaseStreamTestCase.Action(name='action1'),
            BaseStreamTestCase.Action(name='action2'),
        ]
        permissions = {
            'role1': 'rwxcd',
            'role2': 'rw',
        }


    stream_ids = [
        'admin.ru.test_stream',
        'admin.en.test_stream',
        'front.ru.test_stream',
        'front.en.test_stream',
    ]

    mappers = {
        'admin': {
            'ru': {'test_mapper': MagicMock()},
            'en': {'test_mapper': MagicMock()},
        },
        'front': {
            'ru': {'test_mapper': MagicMock()},
            'en': {'test_mapper': MagicMock()},
        },
    }

    def _test_create(self, component):
        super()._test_create(component)
        streams = list(self._get_streams(component))
        for i, db_id in enumerate(self.Stream.db_ids):
            for j, lang in enumerate(self.Stream.langs):
                stream = streams[i*2+j]
                self.assertEqual(stream.lang, lang)
                self.assertEqual(stream.db_id, db_id)
                self.assertEqual(
                    stream.i18n_streams,
                    {strm.lang: strm for strm in streams[i*2:(i+1)*2]},
                )
                self.assertEqual(
                    stream.pub_streams,
                    {strm.db_id: strm for strm in streams[j::2]},
                )
                if not i:
                    self.assertEqual(
                        stream.permissions,
                        {
                            'role1': set('rwxcd'),
                            'role2': set('rw'),
                        }
                    )
                else:
                    self.assertEqual(
                        stream.permissions,
                        {
                            'role1': set('rx'),
                            'role2': set('r'),
                        }
                    )

    @asynctest
    async def test_mapper_prop(self, component):
        for stream in self._get_streams(component):
            test_mapper = self.mappers[stream.db_id][stream.lang]['test_mapper']
            self.assertEqual(stream.mapper, test_mapper)
            self.assertEqual(stream.query(), test_mapper.query())





from iktomi.utils.storage import StorageFrame
from iktomi.utils.storage import VersionedStorage
from iktomi.utils.storage import storage_property
from iktomi.web.route_state import RouteState

class Environment(StorageFrame):

    def __init__(self, app=None, request=None, root=None, _parent_storage=None, **kwargs):
        super(Environment, self).__init__(_parent_storage=_parent_storage, **kwargs)
        self.app = app
        self.request = request
        if self.request:
            self.root = self.app.root.bind_to_env(self._root_storage)
            self._route_state = RouteState(request)
        else:
            self.root = root

    @classmethod
    def create(cls, *args, **kwargs):
        return VersionedStorage(cls, *args, **kwargs)

    def close(self):
        pass

    @storage_property
    def current_location(self):
        ns = getattr(self, 'namespace', '')
        url_name = getattr(self, 'current_url_name', '')
        return '.'.join(filter(None, (ns, url_name)))




import iktomi.web

from .guard import h_guard

h_prefix = iktomi.web.prefix
h_cases = iktomi.web.cases
h_namespace = iktomi.web.namespace


def h_match(path, name=None, convs=None, methods=('GET',), params=()):
    match = iktomi.web.match(path, name=name, convs=convs)
    guard = h_guard(methods, params)
    return  match | guard


class HCases(iktomi.web.cases):

    def __init__(self, *handlers):
        super(HCases, self).__init__(*[h for h in handlers if h])

h_cases = HCases




# coding: utf8
import os

from jinja2 import TemplateNotFound

from ikcms.utils import N_
import ikcms.web
from ikcms.web import h_prefix
from ikcms.web import h_match
from ikcms.web import h_cases
from ikcms.web import h_not_found
from ikcms.web import HView
from ikcms.web import viewhandler


class BaseView(ikcms.web.BaseView):

    def __init__(self, env, data, component, section):
        self.section = component.get_section_with_body(env.db, section['id'])
        if not self.section:
            raise env.app.HTTPNotFound
        self.path = os.path.join(*self.section['path'])
        self.name = self.path
        super(BaseView, self).__init__(env, data)

    @classmethod
    def cases(cls, component, section):
        return []

    @classmethod
    def handler(cls, component, section):
        slug = section['slug']
        return h_prefix('/' + slug, name=slug) | \
            HView(cls, component=component, section=section) | \
            h_cases(*cls.cases(component, section))

    def breadcrumbs(self, children=[]):
        crumbs = [(self.section, self.section.title)] + children
        if self.parent:
            return self.parent.breadcrumbs(crumbs)
        return crumbs

    def url_for_index(self):
        return self.cls_url_for_index(self.root)

    @classmethod
    def cls_url_for_index(cls, root):
        return root.index

    def template_name(self, template):
        if template:
            t = os.path.join(self.path, template)
        else:
            t = self.path

        t = self.env.render.component.resolve(t)
        try:
            self.env.render.component._env.get_template(t)
        except TemplateNotFound:
            if template:
                return os.path.join(self.templates_folder, template)
            else:
                return self.templates_folder
        else:
            return t


class DirView(BaseView):

    name = 'dir'
    title = N_('Dir')

    @classmethod
    def cases(cls, component, section):
        return [
            h_match('/', name='') | h_not_found,
            component.h_subsections(section),
        ]

    @classmethod
    def cls_url_for_index(cls, root):
        return None


class PageView(BaseView):

    name = 'page'
    title = N_('Page')
    templates_folder = 'sections/page'

    @classmethod
    def cases(cls, component, section):
        return [
            h_match('/', name='') | cls.h_index,
            component.h_subsections(section),
        ]

    @viewhandler
    def h_index(self, env, data):
        return self.render_to_response('', dict())




from ikcms.forms import Form

from .. import exceptions


class MessageForm(Form):

    def to_python_or_exc(self, raw_value, keys=None):
        try:
            values, errors = self.to_python(raw_value, keys=keys)
            if errors:
                raise exceptions.MessageError(errors)
        except exceptions.RawValueError as exc:
            raise exceptions.MessageError(
                errors={exc.kwargs['field_name']: exc.kwargs['error']},
            )
        return values





from . import base
import pymemcache

class Component(base.Component):

    DEFAULT_MEMCACHED_HOST = '127.0.0.1'
    DEFAULT_MEMCACHED_PORT = 11211
    prefix = ''

    def __init__(self, app, client):
        super(Component, self).__init__(app)
        self.client = client

    @classmethod
    def create(cls, app):
        host = getattr(app.cfg, 'MEMCACHED_HOST', cls.DEFAULT_MEMCACHED_HOST)
        port = getattr(app.cfg, 'MEMCACHED_PORT', cls.DEFAULT_MEMCACHED_PORT)
        return cls(app, client=pymemcache.Client((host, port)))

    def get(self, key):
        return self.client.get(self._key(key))

    def mget(self, *keys):
        return self.client.get_multi([self._key(key) for key in keys])

    def set(self, key, value, expires=0):
        return self.client.set(self._key(key), value, expire=expires)

    def mset(self, mapping):
        mapping = dict([(self._key(key), value)
                        for key, value in mapping.items()])
        return self.client.set_multi(mapping)

    def add(self, key, value, expires=0):
        return self.client.add(self._key(key), value, expire=expires)

    def delete(self, *keys):
        return self.client.delete_multi([self._key(key) for key in keys])

    def _key(self, key):
        return self.prefix + key


component = Component.create_cls



from ikcms.forms import fields

class Field(fields.Field):

    widget = None

    def __init__(self):
        super().__init__()
        assert self.widget is not None

    def get_cfg(self, env):
        return {
            'name': self.name,
            'label': self.label,
            'widget': self.widget,
        }




import ikcms.components.base
import ikcms.web


class HApp(ikcms.web.WebHandler):

    def __init__(self, component):
        self.component = component
        self.app = component.app

    def app(self, env, data):
        app_request = self.app.get_request(env.request.environ)
        app_env = self.app.get_env(app_request)
        app_env._route_state = env._route_state
        app_env.parent_env = env
        app_data = self.app.get_data()
        return self.app.handler(app_env, app_data)
    __call__ = app


class Component(ikcms.components.base.Component):

    name = None
    App = None
    Cfg = None
    HApp = HApp
    local_cfg = False

    def __init__(self, app):
        assert self.App, 'App property required'
        assert self.Cfg, 'Cfg property required'
        super(Component, self).__init__(app)
        self.app = self.create_app(self.App, self.Cfg)

    def create_app(self, App, Cfg):
        cfg = Cfg(parent_app=self.app, ROOT_DIR=self.app.cfg.ROOT_DIR)
        if self.local_cfg:
            cfg.update_from_py()
        return App(cfg)

    @property
    def root(self):
        return self.app.root

    @property
    def get_env(self):
        return self.app.get_env

    def handler(self):
        return self.HApp(self)


component = Component.create_cls




# -*- coding: utf-8 -*-
from iktomi.cms.menu import MenuGroup, DashRow, DashCol, DashStream, Menu


def top_menu(env):
    return Menu(None, items=[
        Menu(u'Начало', endpoint='index'),
    ], env=env)


def dashboard(env):
    return MenuGroup([
        DashRow([
            DashCol(u'Администрирование', items=[
                DashStream('admins'),
            ]),
        ]),
    ], env=env)



import logging

from . import exceptions
from . import protocols

logger = logging.getLogger(__name__)


class Base:

    def __init__(self, cfg):
        """ Called before started ws server """
        self.cfg = cfg
        self.clients = {}
        self.client_class = self.get_client_class()

    def get_client_class(self):
        from .client import Client
        return Client

    async def __call__(self, server, client_id):
        """ Called when client connected """
        client = await self.add_client(server, client_id)
        while True:
            try:
                raw_request = await server.recv(client_id)
            except Exception as exc:
                await self.handle_connection_error(client, exc)
                break

            request = None
            try:
                request = self.decode_request(raw_request)
                response = await self.handle(client, request)
            except exceptions.ClientError as exc:
                response = await self.handle_client_error(client, exc, request)
            except Exception as exc:
                response = await self.handle_server_error(client, exc, request)

            raw_response = self.encode_response(response)
            try:
                await server.send(client_id, raw_response)
            except Exception as exc:
                await self.handle_connection_error(client, exc)
                break
        await self.remove_client(client_id)

    async def add_client(self, server, client_id):
        if client_id in self.clients:
            raise exceptions.ClientAlreadyAddedError(client_id)
        client = self.client_class(self, server, client_id)
        self.clients[client_id] = client
        return client

    async def remove_client(self, client_id):
        client = self.clients.pop(client_id, None)
        if client is None:
            raise exceptions.ClientNotFoundError(client_id)
        await client.close()

    def decode_request(self, raw_request):
        raise NotImplementedError

    def encode_response(self, raw_request):
        raise NotImplementedError

    async def handle_connection_error(self, client, exc):
        raise NotImplementedError

    async def handle_client_error(self, client, exc, request):
        raise NotImplementedError

    async def handle_server_error(self, client, exc, request):
        raise NotImplementedError


class App(Base):

    protocol = protocols.Json()

    def __init__(self, cfg):
        super().__init__(cfg)
        self.handlers = self.get_handlers()

    def get_handlers(self):
        return {}

    async def handle(self, client, request):
        handler = self.handlers.get(request['handler'])
        if not handler:
            raise exceptions.ClientError(
                exceptions.HandlerNotAllowedError(request['handler']),
            )
        response = await handler(client, request['body'])
        return self.protocol.ResponseMessage(
            name='response',
            request_id=request['request_id'],
            handler=request['handler'],
            body=response,
        )

    async def handle_connection_error(self, client, exc):
        logger.debug(exc, exc_info=True)

    async def handle_client_error(self, client, exc, request):
        logger.debug(exc, exc_info=True)
        request = request or {}
        return self.protocol.ErrorMessage(
            name='error',
            request_id=request.get('request_id', ''),
            handler=request.get('handler', ''),
            body=dict(
                error=exc.error,
                message=str(exc),
                kwargs=exc.kwargs,
            )
        )

    async def handle_server_error(self, client, exc, request):
        logger.exception(exc, exc_info=True)
        request = request or {}
        return self.protocol.ErrorMessage(
            name='error',
            request_id=request.get('request_id'),
            handler=request.get('handler'),
            body=dict(
                error='InternalServerError',
                message='Internal Server Error',
                kwargs={},
            )
        )

    def decode_request(self, raw_request):
        try:
            return self.protocol.decode_request(raw_request)
        except exceptions.ProtocolError as exc:
            raise exceptions.ClientError(exc)

    def encode_response(self, response):
        return self.protocol.encode_response(response)




from unittest import TestCase
from unittest import skipIf
from unittest.mock import MagicMock

import sqlalchemy as sa

from ikcms.ws_components.db import component
from ikcms import orm
from ikcms.orm import relations
from ikcms.utils.asynctests import asynctest
from ikcms.utils.asynctests.db import TableState
from ikcms.utils.asynctests.db import DbState

from tests.cfg import cfg
from tests.models import create_models1
from tests.models import create_metadata


DB_URL = cfg.MYSQL_URL or cfg.POSTGRESS_URL


@skipIf(not cfg.AIO_DB_ENABLED, 'AIO DB DISABLED')
class M2MRelationTestCase(TestCase):

    mapper_cls = orm.mappers.Base

    class LocalMapperClass(mapper_cls):

        name = 'Local'

        def create_columns(self):
            return [
                sa.Column('title', sa.String(255)),
            ]

        def create_relations(self):
            return {
                'm2m': relations.M2M(self, 'Remote'),
                'm2m_ordered': relations.M2M(
                    self,
                    'Remote',
                    ordered=True,
                    tablename="Local_RemoteOrdered"),
            }

    class RemoteMapperClass(mapper_cls):

        name = 'Remote'

        def create_columns(self):
            return [
                sa.Column('title', sa.String(255)),
            ]

    async def asetup(self):
        self.models1 = create_models1()
        registry = orm.mappers.Registry(create_metadata([self.models1]))

        self.LocalMapperClass.create(registry, db_id='db1')
        self.RemoteMapperClass.create(registry, db_id='db1')

        local_mapper = registry['db1']['Local']
        remote_mapper = registry['db1']['Remote']
        registry.create_schema()

        state1 = DbState()
        state1['local'] = TableState(local_mapper.table)
        state1['remote'] = TableState(remote_mapper.table)
        state1['m2m'] = TableState(
            local_mapper.relations['m2m'].table,
            primary_keys=['local_id', 'remote_id'],
        )
        state1['m2m_ordered'] = TableState(
            local_mapper.relations['m2m_ordered'].table,
            primary_keys=['local_id', 'remote_id'],
        )
        state1['remote'].append({'id': 1, 'title': 'title1'})
        state1['remote'].append({'id': 2, 'title': 'title2'})
        state1['remote'].append({'id': 3, 'title': 'title3'})

        state2 = state1.copy()
        state2['local'].append({'id': 1, 'title': 'title1'})
        state2['m2m'].append({'local_id': 1, 'remote_id': 1})
        state2['m2m'].append({'local_id': 1, 'remote_id': 2})
        state2['m2m'].append({'local_id': 1, 'remote_id': 3})
        state2['m2m_ordered'].append({'local_id': 1, 'remote_id': 1, 'order': 2})
        state2['m2m_ordered'].append({'local_id': 1, 'remote_id': 2, 'order': 1})
        state2['m2m_ordered'].append({'local_id': 1, 'remote_id': 3, 'order': 3})

        app = MagicMock()
        del app.db
        app.cfg.DATABASES = {
            'db1': DB_URL,
        }

        Component = component(mappers=registry)

        db = await Component.create(app)
        async with await db() as session:
            conn1 = await session.get_connection(db.engines['db1'])
            await self.models1.reset(conn1)
        return {
            'db': db,
            'local_mapper': local_mapper,
            'remote_mapper': remote_mapper,
            'state1': state1,
            'state2': state2,
        }

    async def aclose(self, db, local_mapper, remote_mapper, state1, state2):
        await db.close()

    @asynctest
    async def test_store(self, db, local_mapper, remote_mapper, state1, state2):
        async with await db() as session:
            await state1.syncdb(session)
            await local_mapper.query().insert_item(
                session,
                {
                    'id': 1,
                    'title': 'title1',
                    'm2m': [1, 2, 3],
                    'm2m_ordered': [2, 1, 3],
                },
            )
            await state2.assert_state(self, session)


    @asynctest
    async def test_load(self, db, local_mapper, remote_mapper, state1, state2):
        async with await db() as session:
            await state2.syncdb(session)
            items = await local_mapper.query().select_items(session)
            self.assertEqual(
                items,
                [{
                    'id': 1,
                    'title': 'title1',
                    'm2m': [1, 2, 3],
                    'm2m_ordered': [2, 1, 3],
                }],
            )
            await state2.assert_state(self, session)


    @asynctest
    async def test_delete(self, db, local_mapper, remote_mapper, state1, state2):
        async with await db() as session:
            await state2.syncdb(session)
            await local_mapper.query().delete_item(session, 1)
            await state1.assert_state(self, session)





from iktomi.cli.sqla import Sqla
from .base import Cli


class DBCli(Cli):

    name = 'DB'

    def __call__(self, *args, **kwargs):
        app = self.create_app(**kwargs)
        metadata_dict = {db_id: models.metadata \
            for db_id, models in app.db.models.items()}
        return Sqla(app.db, metadata_dict, app.db.initial_all)(*args, **kwargs)




from unittest import TestCase
from unittest.mock import MagicMock
from unittest import skipIf

from tests.cfg import cfg

try:
    from ikcms.components.cache.redis import component
    skip_test = False
except ImportError:
    skip_test = True


@skipIf(skip_test, 'Redis not installed')
class RedisTestCase(TestCase):

    def test_cache(self):
        app = self._create_app()

        cache = component().create(app)

        cache.delete(b'test_key')

        value = cache.get(b'test_key')
        self.assertIsNone(value)

        cache.set(b'test_key', b'test_value')
        value = cache.get(b'test_key')
        self.assertEqual(value, b'test_value')

        cache.delete(b'test_key')
        value = cache.get(b'test_key')
        self.assertIsNone(value)

    def _create_app(self):
        app = MagicMock()
        del app.cache
        app.cfg = cfg
        return app



import os
import binascii
from typing import Iterable
from typing import Union
from typing import Any
from typing import Tuple
from iktomi.auth import check_password
from iktomi.auth import encrypt_password
import ikcms.ws_components.base
import ikcms.ws_apps.base.forms

from . import exceptions
from . import mappers
from .forms import message_fields


__all__ = (
    'component',
    'exceptions',
)

def restrict(role=None):
    def wrap(handler):
        async def wrapper(self, client, message):
            if not client.user:
                raise exceptions.ClientError(exceptions.AccessDeniedError())
            return await handler(self, client, message)
        wrapper.handler = handler
        return wrapper
    return wrap


class AuthForm(ikcms.ws_apps.base.forms.MessageForm):
    fields = [
        message_fields.token,
        message_fields.login,
        message_fields.password,
    ]


class Component(ikcms.ws_components.base.Component):
    name = 'auth'
    requirements = ['db', 'cache']
    users_mapper = 'main.AdminUser'

    def client_init(self, client):
        client.user = None

    async def h_login(self, client, message):
        form = AuthForm()
        try:
            data = form.to_python_or_exc(message)
        except MessageError as exc:
            raise exceptions.ClientError(exc)
        token = data.get('token', None)
        login = data.get('login', None)
        password = data.get('password', None)
        if token is not None:
            try:
                token = await self.login_by_token(client, token)
            except exceptions.InvalidTokenError as exc:
                raise exceptions.ClientError(exc)
        elif login is not None and password is not None:
            try:
                token = await self.login_by_password(client, login, password)
            except exceptions.InvalidPasswordError as exc:
                raise exceptions.ClientError(exc)
        else:
            raise exceptions.ClientError(exceptions.InvalidCredentialsError())
        return {
            'user': {
                'login': client.user['login'],
                'name': client.user['name'],
            },
            'token': token,
        }

    async def h_logout(self, client, message):
        try:
            await self.logout(client)
        except exceptions.AccessDeniedError as exc:
            raise exceptions.ClientError(exc)
        return {}

    async def login_by_token(self, client, token):
        user, token = await self.auth_by_token(token)
        client.user = user
        return token

    async def login_by_password(self, client, login, password):
        user, token = await self.auth_by_password(login, password)
        client.user = user
        return token

    async def logout(self, client):
        if client.user is None:
            raise exceptions.AccessDeniedError
        client.user = None

    async def get_user_by_login(self, login):
        users_mapper = self.app.db.mappers.get_mapper(self.users_mapper)
        query = users_mapper.query().filter_by(login=login)
        async with await self.app.db() as session:
            user = await query.select_first_item(session)
            if user:
                await users_mapper.relations['groups'].m.query().fill(
                    session, user, 'groups')
        return user

    async def get_user_by_token(self, token):
        login = await self.app.cache.get(token)
        if login is not None:
            return await self.get_user_by_login(login)
        return None

    async def auth_by_token(self, token):
        user = await self.get_user_by_token(token)
        if user is None:
            raise exceptions.InvalidTokenError()
        return user, token

    async def auth_by_password(self, login, password):
        user = await self.get_user_by_login(login)
        if user is None or not check_password(password, user['password']):
            raise exceptions.InvalidPasswordError()
        token = binascii.hexlify(os.urandom(10)).decode('ascii')
        await self.app.cache.set(token, login)
        return user, token

    def get_user_roles(self, user):
        roles = set()
        for group in user['groups']:
            print(group)
            roles.update(group['roles'])
        return roles

    def get_user_perms(self, user, permissions):
        perms = set()
        roles = self.get_user_roles(user)
        for role in roles:
            perms.update(permissions.get('role', []))
        return list(perms)

    def check_perms(self, user, perms):
        user_perms = self.component.app.auth.\
            get_user_perms(user, self.permissions)
        if not set(perms).issubset(user_perms):
            raise exceptions.AccessDeniedError


component = Component.create_cls



from iktomi.utils import cached_property
from iktomi.db.files import FileManager
from iktomi.cms.packer import StaticPacker

import ikcms.apps.composite
import ikcms.components.cache.redis
import ikcms.components.db.cli
import ikcms.components.i18n
import ikcms.cli.app
import ikcms.cli.db

from . import components


class App(ikcms.apps.composite.App):

    components = [
        ikcms.components.cache.redis.component(),
        components.db.component(),
        components.render.component(
            paths=[
                'pkg://iktomi.cms/templates',
                'pkg://iktomi/templates/jinja2/templates',
            ],
        ),
        ikcms.components.i18n.component(),
    ]

    commands = {
        'admin': ikcms.cli.app.AppCli,
        'db': ikcms.cli.db.DBCli,
        'generator': ikcms.components.db.cli.GeneratorCli,
    }

    def __init__(self, cfg):
        self.streams = self.get_streams()
        super(App, self).__init__(cfg)

    def get_handler(self):
        from .handler import get_handler
        return get_handler(self)

    def get_env_class(self):
        from .env import Environment
        return Environment

    @cached_property
    def admin_file_manager(self):
        return FileManager(
            self.cfg.FORM_TEMP,
            self.cfg.MEDIA_ROOT_ADMIN,
            self.cfg.FORM_TEMP_URL,
            self.cfg.MEDIA_URL,
        )

    @cached_property
    def front_file_manager(self):
        return FileManager(
            transient_root=None,
            persistent_root=self.cfg.MEDIA_ROOT_FRONT,
            transient_url=None,
            persistent_url=self.cfg.MEDIA_URL,
        )

    def get_streams(self):
        from .streams import streams
        return streams

    def get_dashboard(self, env):
        from .menuconf import dashboard
        return dashboard(env)

    def get_top_menu(self, env):
        from .menuconf import top_menu
        return top_menu(env)

    def get_context(self, env):
        from .context import Context
        return Context(env)

    @cached_property
    def packer(self):
        return StaticPacker()




import logging

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.query import Query

import ikcms.components.db.base


class Component(ikcms.components.db.base.DBComponent):

    session_maker_class = sessionmaker
    query_class = Query
    generator = None

    def __init__(self, app, engines, models):
        super(Component, self).__init__(app)
        self.engines = engines
        self.models = models
        self.binds = self.get_binds()
        self.session_maker = self.create_session_maker()

    def create_session_maker(self):
        return self.session_maker_class(
            binds=self.binds,
            query_cls=self.query_class,
        )

    @classmethod
    def create(cls, app):
        databases = getattr(app.cfg, 'DATABASES', {})
        database_params = getattr(app.cfg, 'DATABASE_PARAMS', {})
        engines = {}
        for db_id, url in databases.items():
            engine = cls.create_engine(db_id, url, database_params)
            engines[db_id] = engine
            engine.logger = logging.getLogger('sqlalchemy.engine.[%s]' % db_id)
        models = {db_id: cls.get_models(db_id) for db_id in databases}
        return cls(app, engines, models)

    @classmethod
    def create_engine(cls, db_id, url, engine_params):
        engine = create_engine(url, **engine_params)
        engine.db_id = db_id
        return engine

    def on_initialization_end(self):
        self.close()

    def on_init_env(self, env):
        setattr(env, self.name, self())
        #env.models = self.env_models

    def on_close_env(self, env):
        getattr(env, self.name).close()

    @staticmethod
    def get_models(db_id):
        import models
        return getattr(models, db_id)

    def get_binds(self):
        binds = {}
        for db_id, engine in self.engines.items():
            for table in self.models[db_id].metadata.sorted_tables:
                binds[table] = engine
        return binds

    def __call__(self):
        return self.session_maker()

    def close(self):
        for engine in self.engines.values():
            engine.dispose()

    def create_all(self):
        for db_id, models in self.models.items():
            models.metadata.create_all(self.engines[db_id])

    def initial_all(self, session):
        from models.initial import initialize
        initialize(self.app, session)

    def drop_all(self):
        # XXX Need confirmation?
        for db_id, models in self.models.items():
            models.metadata.drop_all(self.engines[db_id])

    def reset_all(self):
        self.drop_all()
        self.create_all()

    def get_model(self, path):
        paths = path.split('.')
        try:
            model = self.models[paths[0]]
            for p in paths[1:]:
                model = getattr(model, p)
        except (KeyError, AttributeError):
            raise ValueError(path)
        return model

    def generate(self, names=None):
        if self.generator:
            self.generator(self, names)
        else:
            raise NotImplementedError

component = Component.create_cls





from iktomi.web import WebHandler
from iktomi.web import namespace
from iktomi.web.reverse import Location


class LanguageLocation(Location):
    def __init__(self, *builders, **kwargs):
        self.langs = kwargs.pop('langs')
        Location.__init__(self, *builders, **kwargs)


class HLang(WebHandler):

    def __init__(self, component, active):
        assert active in component.langs
        self.component = component
        self.active = active
        self._next_handler = namespace(active)

    def __call__(self, env, data):
        self.component.set_lang(env, self.active)
        return self.next_handler(env, data)

    def _filter_locations(self, _locations):
        return dict([(nm, (loc, self._filter_locations(scope)))
                     for (nm, (loc, scope)) in _locations.items()
                     if not isinstance(loc, LanguageLocation)
                     or self.active in loc.langs])

    def _locations(self):
        locations = WebHandler._locations(self)
        return self._filter_locations(locations)


class HForLangs(WebHandler):
    def __init__(self, component, *langs):
        self.component = component
        self.langs = frozenset(langs)

    def __call__(self, env, data):
        if env.lang in self.langs:
            return self.next_handler(env, data)

    def _locations(self):
        locations = WebHandler._locations(self)
        new_locations = {}
        for name, (loc, scope) in locations.items():
            cls = loc.__class__
            assert cls in [Location], \
                'for_languages sublocations class should be exact Location ' \
                'otherwise we can not be shure that location class change ' \
                'brakes nothing %s' % loc.__class__.__name__
            loc = LanguageLocation(*loc.builders, langs=self.langs)
            new_locations[name] = (loc, scope)
        return new_locations




import sys
from lxml import etree

SPHINX_NAMESPACE = 'http://sphinxsearch.com/'

_ns = '{' + SPHINX_NAMESPACE + '}'


class BaseSlot(object):

    def __init__(self, name):
        self.name = name

    def convert(self, value):
        return unicode(value)


class Field(BaseSlot):

    def write_schema(self, xf):
        with xf.element(_ns + 'field', name=self.name):
            pass


class BaseAttr(BaseSlot):

    type = None

    def __init__(self, name, default=None):
        BaseSlot.__init__(self, name)
        self.default = default

    def _schema_attrib(self):
        attrs = {'name': self.name, 'type': self.type}
        if self.default is not None:
            attrs['default'] = self.convert(self.default)
        return attrs

    def write_schema(self, xf):
        with xf.element(_ns + 'attr', attrib=self._schema_attrib()):
            pass


class Int(BaseAttr):

    type = 'int'
    bits = None

    def __init__(self, name, bits=32, default=None):
        BaseAttr.__init__(self, name, default=default)
        assert bits in xrange(1, 33)
        self.bits = int(bits)

    def _schema_attrib(self):
        attrs = BaseAttr._schema_attrib(self)
        attrs['bits'] = str(self.bits)
        return attrs


class BigInt(BaseAttr):

    type = 'bigint'


class String(BaseAttr):

    type = 'string'


class Bool(BaseAttr):

    type = 'bool'

    def convert(self, value):
        return '1' if value else '0'


class Float(BaseAttr):

    type = 'float'


class Multi(BaseAttr):
    '''List of int32'''

    type = 'multi'

    def convert(self, value):
        return ' '.join(str(item) for item in value)


class Timestamp(BaseAttr):

    type = 'timestamp'

    def convert(self, value):
        return value.strftime('%s')


class Pipe(object):

    def __init__(self, fields, attrs=[], fp=sys.stdout):
        self.slots = {s.name: s for s in [Field(n) for n in fields] + attrs}
        self._xf_cm = etree.xmlfile(fp, encoding='utf-8')

    def __enter__(self):
        self._xf = xf = self._xf_cm.__enter__()
        xf.write_declaration()
        self._root = xf.element(_ns + 'docset',
                                nsmap={'sphinx': SPHINX_NAMESPACE})
        self._root.__enter__()
        self.write_schema()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._root.__exit__(exc_type, exc_value, exc_tb)
        self._xf_cm.__exit__(exc_type, exc_value, exc_tb)

    def write_schema(self):
        with self._xf.element(_ns + 'schema'):
            for slot in self.slots.values():
                slot.write_schema(self._xf)

    def document(self, id, **data):
        with self._xf.element(_ns + 'document', id=str(id)):
            for name, value in data.items():
                slot = self.slots[name]
                value = slot.convert(value)
                with self._xf.element(name):
                    self._xf.write(value)

    def killlist(self, ids):
        with self._xf.element(_ns + 'killlist'):
            for id in ids:
                with self._xf.element('id'):
                    self._xf.write(str(id))



import sqlalchemy as sa
from iktomi.utils import cached_property

__all__ = (
    'M2M',
    'I18nM2M',
)

class M2M:

    order_field_name = 'order'

    def __init__(self, mapper1, mapper_name2, ordered=False, tablename=None):
        self.mapper1 = mapper1
        self.registry = mapper1.registry
        self.db_id = mapper1.db_id
        self.mapper_name1 = mapper1.name
        self.mapper_name2 = mapper_name2
        self.ordered = ordered
        self.tablename = tablename or self.get_tablename()

    def create_tables(self):
        columns = [
            sa.Column(
                self.local_field_name,
                self.mapper1.c['id'].type,
                sa.ForeignKey(self.mapper1.table.c.id, ondelete="CASCADE"),
                nullable=False,
            ),
            sa.Column(
                self.remote_field_name,
                self.mapper2.c['id'].type,
                sa.ForeignKey(self.mapper2.table.c.id, ondelete="CASCADE"),
                nullable=False,
            ),
        ]
        if self.ordered:
            columns.append(
                sa.Column('order', sa.Integer, nullable=False, default=0)
            )
        return sa.Table(
            self.tablename,
            self.registry.metadata[self.db_id],
            *columns
        )

    def get_tablename(self):
        return "{}_{}".format(self.mapper_name1, self.mapper_name2)

    @cached_property
    def table(self):
        return self.registry.metadata[self.db_id].tables[self.tablename]

    @cached_property
    def mapper2(self):
        return self.registry[self.db_id][self.mapper_name2]

    @cached_property
    def m(self):
        return self.mapper2

    @cached_property
    def local_field_name(self):
        return '{}_id'.format(self.mapper1.name.lower())

    @cached_property
    def remote_field_name(self):
        return '{}_id'.format(self.mapper2.name.lower())

    @cached_property
    def local_field(self):
        return self.table.c[self.local_field_name]

    @cached_property
    def remote_field(self):
        return self.table.c[self.remote_field_name]

    @cached_property
    def order_field(self):
        return self.table.c[self.order_field_name]

    async def load(self, session, item_ids):
        result = {}
        for row in await session.execute(self.select_query(item_ids)):
            result.setdefault(row[self.local_field], []) \
                .append(row[self.remote_field])
        return result

    async def store(self, session, item_id, value):
        await session.execute(self.delete_query(item_id))
        for num, rel_id in enumerate(value):
            order = self.ordered and (num+1) or None
            await session.execute(
                self.insert_query(item_id, rel_id, order=order))

    async def delete(self, session, item_id):
        await session.execute(self.delete_query(item_id))

    def select_query(self, item_ids):
        s = sa.sql.select([self.local_field, self.remote_field]) \
            .where(self.local_field.in_(item_ids))
        if self.ordered:
            s = s.order_by(self.order_field)
        return s

    def delete_query(self, local_id):
        return sa.sql.delete(self.table).where(self.local_field == local_id)

    def insert_query(self, local_id, remote_id, order=None):
        values = {
            self.local_field_name: local_id,
            self.remote_field_name: remote_id,
        }
        if order:
            values[self.order_field_name] = order
        return sa.sql.insert(self.table).values(**values)


class I18nM2M(M2M):

    def __init__(self, mapper1, mapper_name2, ordered=False, tablename=None):
        super().__init__(
            mapper1,
            mapper_name2,
            ordered=ordered,
            tablename=tablename,
        )
        self.lang = mapper1.lang

    @property
    def get_tablename(self):
        return "{}_{}{}".format(
            self.mapper_name1, self.mapper_name2, self.lang.capitalize())

    @property
    def mapper1(self):
        return self.registry[self.db_id][self.lang][self.mapper_name1]

    @property
    def mapper2(self):
        return self.registry[self.db_id][self.lang][self.mapper_name2]





from ikcms.web import (
    h_cases,
    h_match,
    h_static_files,
    h_404,
)


def h_index(env, data):
    return env.render.to_response('index.html')


def get_handler(app):
    return h_cases(
        h_static_files(app.cfg.STATIC_DIR, app.cfg.STATIC_URL),
        h_static_files(app.cfg.MEDIA_DIR, app.cfg.MEDIA_URL),
        h_match('/', name='index') | h_index,
        h_404,
    )



import ikcms.ws_apps.base.client


class Client(ikcms.ws_apps.base.client.Client):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for component in self.app.components:
            component.client_init(self)

    async def close(self):
        for component in self.app.components:
            component.client_close(self)




import logging
from hashlib import md5
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO
from .. import Request
from .. import Response
from .. import WebHandler
from .. import exceptions
from ..handlers import request_filter


logger = logging.getLogger(__name__)


class CacheableRequestError(Exception):
    pass


class CacheableRequest(object):
    HEADERS_ACCESS_ERROR = 'Could not access headers on cacheable request'
    COOKIES_ACCESS_ERROR = 'Could not access cookies on cacheable request'
    GET_ACCESS_ERROR = 'Could not access GET params on cacheable request'
    SETATTR_ERROR = 'Could not modify cacheable request'

    def __init__(self, request):
        if isinstance(request, CacheableRequest):
            request = request.unwrap()
        object.__setattr__(self, 'request', request)

    def unwrap(self):
        return self.request

    @property
    def headers(self):
        raise CacheableRequestError(self.HEADERS_ACCESS_ERROR)

    @property
    def cookies(self):
        raise CacheableRequestError(self.COOKIES_ACCESS_ERROR)

    @property
    def GET(self):
        raise CacheableRequestError(self.GET_ACCESS_ERROR)

    def __getattr__(self, item):
        return getattr(self.request, item)

    def __setattr__(self, key, value):
        raise CacheableRequestError(self.SETATTR_ERROR)


class HCache(WebHandler):
    CFG_ENABLED = 'CACHE_RESPONSE_ENABLED'
    CFG_EXPIRES = 'CACHE_RESPONSE_EXPIRES'
    CFG_BACKEND = 'CACHE_RESPONSE_BACKEND'

    def __init__(self, cfg):
        super(HCache, self).__init__()
        self.enabled = getattr(cfg, self.CFG_ENABLED, False)
        self.expires = getattr(cfg, self.CFG_EXPIRES, 0)
        self.backend = BACKENDS[getattr(cfg, self.CFG_BACKEND, 'redis')]()
        self.nocache = request_filter(
            lambda env, data, next_handler:
                self.backend.nocache(env, data, next_handler)
        )

    def should_cache_response(self, env):
        if not self.enabled:
            return False
        if not env.request.method in ['GET', 'HEAD']:
            return False
        return True

    def try_cache(self, env, data):
        if self.should_cache_response(env):
            env.request = CacheableRequest(env.request)
            try:
                response = self.backend.try_cache(
                    env, data, self.next_handler, self.expires,
                )
            finally:
                env.request = env.request.unwrap()
            return response
        return self.next_handler(env, data)

    __call__ = try_cache


class NginxBackend(object):
    def should_cache_response(self, response):
        if not isinstance(response, Response):
            return False
        if not response.status_code // 100 == 2:
            return False
        if 'X-Accel-Expires' in response.headers:
            return False
        return True

    def try_cache(self, env, data, next_handler, expires):
        response = next_handler(env, data)
        if self.should_cache_response(response):
            response.headers['X-Accel-Expires'] = str(expires)
        return response

    def nocache(self, env, data, next_handler):
        if isinstance(env.request, CacheableRequest):
            env.request = env.request.unwrap()
        response = next_handler(env, data)
        if isinstance(response, Response):
            response.headers['X-Accel-Expires'] = '0'
        return response


class RedisBackend(object):
    NOCACHE_ATTR = 'CACHE_RESPONSE_NOCACHE_ATTR'

    def should_cache_response(self, response):
        if not isinstance(response, Response):
            return False
        if not response.status_code // 100 == 2:
            return False
        if hasattr(response, self.NOCACHE_ATTR):
            return False
        return True

    def try_cache(self, env, data, next_handler, expires):
        key = 'CACHE_RESPONSE_' + md5(env.request.url).hexdigest()
        response = env.app.cache.get(key)
        if response is None:
            response = next_handler(env, data)
            if self.should_cache_response(response):
                env.app.cache.set(key, str(response), expires=expires)
                logger.info('Put response to cache for {} sec'.format(expires))
            else:
                logger.info('Skip response cache')
        else:
            response = Response.from_file(StringIO(response))
            logger.info('Get response from cache')
        return response


    def nocache(self, env, data, next_handler):
        if isinstance(env.request, CacheableRequest):
            env.request = env.request.unwrap()
        response = next_handler(env, data)
        if isinstance(response, Response):
            setattr(response, self.NOCACHE_ATTR, self.NOCACHE_ATTR)
        return response


BACKENDS = {
    'redis': RedisBackend,
    'nginx': NginxBackend,
}



import lxml.html
import lxml.etree
from jinja2 import Markup

import ikcms.components.base

from .model import ExpandableMarkup


class Component(ikcms.components.base.Component):

    name = 'markup'

    # replacements = [(u'<p>pattern</p>', u'<h1>replacement</h1>')]
    replacements = []

    # def open_new_window_filter(root, **kwargs):
    #     for tag in root.xpath('//a'):
    #         tag.attrib['target'] = '_blank'
    #
    # filters = [open_new_window_filter]
    filters = []

    def apply_replacements(self, value):
        for pattern, replacement in self.replacements:
            if not isinstance(pattern, Markup):
                pattern = Markup(pattern)
            if not isinstance(replacement, Markup):
                replacement = Markup(replacement)
            value = value.replace(pattern, replacement)
        return value

    def apply_filters(self, value, **kwargs):
        try:
            root = lxml.html.fragment_fromstring(value, create_parent=True)
            for filter in self.filters:
                filter(self, root, **kwargs)
        except lxml.etree.XMLSyntaxError as e:
            return value
        else:
            return self.stringify(root)

    def expand(self, value, **kwargs):
        if isinstance(value, ExpandableMarkup):
            str_value = value.markup
            str_value = self.apply_filters(str_value, **kwargs)
            str_value = self.apply_replacements(str_value)
            return Markup(str_value)
        return value

    @staticmethod
    def stringify(tag, encoding='utf-8'):
        head = tag.text or ''
        tail = ''.join([
            lxml.html.tostring(child, encoding=encoding, method='xml')
            for child in tag.iterchildren()
        ])
        return ''.join([head, tail.decode(encoding)])


component = Component.create_cls



from __future__ import absolute_import
import six
import time
import random

import redis

from . import base


class Lock(object):

    def __init__(self, component, key, expires=60, timeout=10, sleep=0.5):
        self.component = component
        self.key = key
        self.timeout = timeout
        self.expires = expires
        self.sleep = sleep
        self.lock_id = str(random.randint(1, 1000000))

    def __enter__(self):
        timeout = self.timeout
        while timeout >= 0:

            if self.component.add(self.key, self.lock_id, self.expires):
                return

            timeout -= self.sleep
            time.sleep(self.sleep)

        raise self.component.LockTimeout("Timeout whilst waiting for lock")

    def __exit__(self, exc_type, exc_value, traceback):
        with self.component.pipe() as pipe:
            try:
                pipe.watch(self.key)
                lock_id = pipe.get(self.key)
                if lock_id == self.lock_id:
                    pipe.execute()
                    pipe.delete(self.key)
                else:
                    raise self.component.LockLosted(self.key)
            except self.component.WatchError:
                raise self.component.LockLosted(self.key)


class Pipe(object):

    def __init__(self, component, pipe):
        self._pipe = pipe

    def get(self, key):
        result = self._pipe.get(key)
        if isinstance(result, six.binary_type):
            return result.decode()
        return result

    def mget(self, *keys):
        return self._pipe.mget(*keys)

    def set(self, key, value, expires=None):
        return self._pipe.set(key, value, ex=expires)

    def mset(self, mapping):
        return self._pipe.mset(mapping)

    def hget(self, key, hkey):
        return self._pipe.hget(key, hkey)

    def hset(self, key, hkey, value):
        return self._pipe.hset(key, hkey, value)

    def hmget(self, key, hkeys):
        return self._pipe.hmget(key, hkeys)

    def hmset(self, key, mapping):
        return self._pipe.hmset(key, mapping)

    def hkeys(self, key):
        return self._pipe.hkeys(key)

    def hvals(self, key):
        return self._pipe.hvals(key)

    def hdel(self, key, *hkeys):
        return self._pipe.hdel(key, *hkeys)

    def add(self, key, value, expires=0):
        return self._pipe.set(key, value, ex=expires, nx=True)

    def delete(self, *keys):
        return self._pipe.delete(*keys)

    def watch(self, key):
        return self._pipe.watch(key)

    def execute(self):
        return self._pipe.execute()

    def __enter__(self):
        self._pipe.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self._pipe.__exit__(exc_type, exc_value, traceback)


class Component(base.Component):

    DEFAULT_REDIS_URL = "redis://localhost:6379/0"
    WatchError = redis.WatchError

    class LockTimeout(Exception): pass
    class LockLosted(Exception): pass

    def __init__(self, app, client):
        super(Component, self).__init__(app)
        self.client = client

    @classmethod
    def create(cls, app):
        url = getattr(app.cfg, 'REDIS_URL', cls.DEFAULT_REDIS_URL)
        return cls(app, client=redis.StrictRedis.from_url(url))

    def get(self, key):
        result = self.client.get(key)
        if isinstance(result, six.binary_type):
            return result.decode()
        return result

    def mget(self, *keys):
        return self.client.mget(*keys)

    def set(self, key, value, expires=None):
        return self.client.set(key, value, ex=expires)

    def mset(self, mapping):
        return self.client.mset(mapping)

    def add(self, key, value, expires=None):
        return self.client.set(key, value, ex=expires, nx=True)

    def delete(self, *keys):
        return self.client.delete(*keys)

    def hget(self, key, hkey):
        return self.client.hget(key, hkey)

    def hset(self, key, hkey, value):
        return self.client.hset(key, hkey, value)

    def hmget(self, key, hkeys):
        return self.client.hmget(key, hkeys)

    def hmset(self, key, mapping):
        return self.client.hmset(key, mapping)

    def hkeys(self, key):
        return self.client.hkeys(key)

    def hvals(self, key):
        return self.client.hvals(key)

    def hdel(self, key, *hkeys):
        return self.client.hdel(key, *hkeys)

    def pipe(self):
        return Pipe(self, self.client.pipeline())

    def lock(self, key, expires=60, timeout=10, sleep=0.5):
        return Lock(self, key, expires, timeout, sleep)


component = Component.create_cls



import six
from iktomi.cms.auth.views import AdminAuth, auth_required
from iktomi.cms.item_lock.views import ItemLockView
from iktomi.cms.menu import IndexHandler

from ikcms.web import h_match
from ikcms.web import h_prefix
from ikcms.web import h_cases
from ikcms.web import h_static_files
from ikcms.web import h_app
from ikcms.web import h_domain

__all__ = ['get_handler']


def get_handler(app):
    admin_models = app.db.get_models('admin')
    auth = AdminAuth(admin_models.AdminUser, app.cache, expire_time=None)

    if getattr(app, 'preview', False):
        h_preview = app.preview.handler()
    else:
        h_preview = None


    return h_domain(default=getattr('app.cfg', 'DEFAULT_DOMAIN', None)) | \
        h_cases(
            h_static_files(
                app.cfg.STATIC_DIR,
                app.cfg.STATIC_URL,
                app.cfg.STATIC_ENABLED,
            ),
            h_static_files(
                app.cfg.MEDIA_ROOT_ADMIN,
                app.cfg.MEDIA_URL,
                app.cfg.STATIC_ENABLED,
            ),
            h_static_files(
                app.cfg.CMS_STATIC_DIR,
                app.cfg.CMS_STATIC_URL,
                app.cfg.STATIC_ENABLED,
            ),

            h_match('/pack.js', name='js_packer') | app.packer.js_packer,
            h_match('/pack.css', name='css_packer') | app.packer.css_packer,
            auth.login(),
            auth.logout(),

            auth | auth_required | h_cases(
                h_preview,
                h_match('/', name='index', params=(('__ajax', six.u),)) |
                    IndexHandler(app.get_dashboard),
                ItemLockView().app,
                app.streams.to_app(),
            ),
        )



# -*- coding: utf-8 -*-
import logging
from jinja2 import nodes
import jinja2.ext
from jinja2.exceptions import TemplateSyntaxError


__all__ = (
    'ShowTag',
    'CacheTag',
    'I18n',
    'Macros',
)

logger = logging.getLogger(__name__)

class ShowTag(jinja2.ext.Extension):
    allowed_languages = ['ru', 'en']

    tags = set(['show'])

    def parse(self, parser):
        lineno = parser.stream.next().lineno
        token = next(parser.stream)

        if token.value not in self.allowed_languages:
            raise TemplateSyntaxError('Expected language token from set: %s' %
                                      ', '.join(self.allowed_languages), lineno)

        body = parser.parse_statements(['name:endshow'], drop_needle=True)
        call_method = self.call_method(
            '_show_support',
            [nodes.ContextReference(), nodes.Const(token.value)],
        )
        node = nodes.CallBlock(call_method, [], [], body).set_lineno(lineno)
        return node

    def _show_support(self, context, lang, caller):
        if context['env'].lang == lang:
            return caller()
        return u''


class CacheTag(jinja2.ext.Extension):
    # a set of names that trigger the extension.
    tags = set(['cache'])

    cache_key_prefix = 'jinja2_cached_block_'

    def __init__(self, environment):
        super(CacheTag, self).__init__(environment)

        # add the defaults to the environment
        environment.extend(
            fragment_cache_prefix='',
            fragment_cache=None
        )

    def parse(self, parser):
        # the first token is the token that started the tag.  In our case
        # we only listen to ``'cache'`` so this will be a name token with
        # `cache` as value.  We get the line number so that we can give
        # that line number to the nodes we create by hand.
        lineno = parser.stream.next().lineno

        # now we parse a single expression that is used as cache key.
        args = [parser.parse_expression()]

        # if there is a comma, the user provided a timeout.  If not use
        # None as second parameter.
        if parser.stream.skip_if('comma'):
            args.append(parser.parse_expression())
        else:
            args.append(nodes.Const(None))

        # now we parse the body of the cache block up to `endcache` and
        # drop the needle (which would always be `endcache` in that case)
        body = parser.parse_statements(['name:endcache'], drop_needle=True)

        # now return a `CallBlock` node that calls our _cache_support
        # helper method on this extension.
        call_method = self.call_method(
            '_cache_support',
            [ContextReference()] + args,
        ),
        return nodes.CallBlock(call_method, [], [], body).set_lineno(lineno)

    def _cache_support(self, context, name, timeout=None, caller=None):
        """Helper callback."""
        #enabled = getattr(self.environment.cfg, 'CACHE_PAGES_ENABLED', True)
        env = context['env']

        enabled = env.cfg.CACHE_BLOCKS_ONLY or env.cfg.CACHE_ENABLED
        if not enabled:
            return caller()

        if not timeout:
            timeout = getattr(env.cfg, 'CACHE_BLOCKS_TIME', 60)

        key = '_'.join([self.cache_key_prefix, name, env.lang])

        # try to load the block from the cache
        # if there is no fragment in the cache, render it and store
        # it in the cache.
        rv = env.cache.get(key)
        if rv is not None:
            return rv
        rv = caller()
        env.cache.set(key, rv, timeout)
        return rv


class I18n(jinja2.ext.i18n):

    def __init__(self, environment):
        super(I18n, self).__init__(environment)
        environment.install_null_translations()


class MacrosModuleWrapper(object):

    def __init__(self, environment, name):
        self.__environment = environment
        self.__name = name

    def __getattr__(self, name):
        template = self.__environment.get_template(self.__name)
        module = template.make_module(vars=self.__environment.globals)
        return getattr(module, name)

    def __call__(self, *args, **kwargs):
        return self.main(*args, **kwargs)


class MacrosLib(object):

    def __init__(self, environment):
        self.environment = environment

    def __getattr__(self, name):
        try:
            result = MacrosModuleWrapper(
                self.environment,
                'macros/{}.html'.format(name)
            )
            setattr(self, name, result)
            return result
        except Exception as e:
            logger.exception(e)
            raise


class Macros(jinja2.ext.Extension):

    def __init__(self, environment):
        environment.globals['macros'] = MacrosLib(environment)



from unittest import TestCase
from unittest import skipIf
from unittest.mock import MagicMock

from ikcms.utils.asynctests import asynctest

from tests.cfg import cfg

if cfg.AIOMCACHE_ENABLED:
    from ikcms.ws_components.cache.aiomcache import component



@skipIf(not cfg.AIOMCACHE_ENABLED, 'Aiomcache DISABLED')
class AIOMCacheTestCase(TestCase):

    @asynctest
    async def test_cache(self):
        app = self._create_app()

        cache = await component().create(app)
        await cache.delete(b'test_key')

        value = await cache.get(b'test_key')
        self.assertIsNone(value)

        await cache.set(b'test_key', b'test_value')
        value = await cache.get(b'test_key')
        self.assertEqual(value, b'test_value')

        await cache.delete(b'test_key')
        value = await cache.get(b'test_key')
        self.assertIsNone(value)

        app = self._create_app()

        cache_with_key = await component(prefix=b'test_prefix-').create(app)
        await cache.delete(b'test_prefix-test_key')
        await cache_with_key.set(b'test_key', b'test_value')

        value = await cache.get(b'test_prefix-test_key')
        self.assertEqual(value, b'test_value')
        value = await cache_with_key.get(b'test_key')
        self.assertEqual(value, b'test_value')
        await cache_with_key.delete(b'test_key')
        value = await cache.get(b'test_prefix-test_key')
        self.assertIsNone(value)

    def _create_app(self):
        app = MagicMock()
        del app.cache
        app.cfg = cfg
        return app



from json import dumps
from webob import Response
from iktomi.templates import BoundTemplate as BaseBoundTemplate
from ikcms.components.base import Component


class BoundTemplate(BaseBoundTemplate):

    def __init__(self, component, env):
        self.component = component
        self.env = env

    def render(self, template_name, **context):
        return self.component(template_name, **self.get_template_vars(context))
    __call__ = render


    def to_response(self, template_name, context=None, content_type=None):
        context = context or {}
        return self.component.to_response(
            template_name,
            self.get_template_vars(context),
            content_type=content_type
        )

    def to_json(self, *args, **kwargs):
        return self.component.to_json(*args, **kwargs)

    def get_template_vars(self, context):
        storage = self.env._root_storage
        vs = {'env': storage}
        if hasattr(storage, 'get_template_vars'):
            vs.update(storage.get_template_vars())
        vs.update(context)
        return vs


class RenderComponent(Component):

    name = 'render'
    env_component_class = BoundTemplate

    def on_init_env(self, env):
        setattr(env, self.name, self.env_component_class(self, env))

    def render(self, template_name, **context):
        raise NotImplementedError
    __call__ = render

    def to_response(self, name, context=None, content_type='text/html'):
        context = context or {}
        result = self.render(name, **context)
        return Response(result, content_type=content_type)

    def to_json(self, context, content_type='application/json'):
        return Response(
            dumps(context),
            content_type=content_type,
            charset='UTF-8',
        )



from ikcms.forms import fields


__all__ = (
    'String',
    'Int',
    'Find',
    'Date',
    'id',
    'title',
)


def eq_filter(field, query, value):
    column = field.context['stream'].mapper.table.c[field.name]
    if value is not None:
        query = query.where(column == value)
    return query

def like_filter(field, query, value):
    column = field.context['stream'].mapper.table.c[field.name]
    if value is not None:
        query = query.where(column.like("%{}%".format(value)))
    return query


class String(fields.String):
    filter = eq_filter
    raw_required = False


class Int(fields.Int):
    filter = eq_filter
    raw_required = False


class Find(String):
    filter = like_filter
    raw_required = False


class Date(fields.Date):
    filter = eq_filter #XXX Filter mast been between
    raw_required = False


class id(Int):
    name = 'id'
    label = 'Id'


class title(Find):
    name = 'title'
    label = 'Title'





from ikcms.ws_apps.base.exceptions import BaseError


__all__ = (
    'ComponentNotFoundError',
)

class ComponentNotFoundError(BaseError):
    message = 'Component not found: {component}'

    def __init__(self, component):
        self.kwargs = {
            'component': component,
        }

